from typing import Any
import lightning as L
import torch.nn as nn
from lightning.pytorch.utilities.model_summary import ModelSummary


class Model(L.LightningModule):
    def __init__(self, model, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.model = model


def convert_int_to_shortened_string(num):
    """
    Converts an integer into a shortened format using suffixes like K, M, B, and T for
    thousands, millions, billions, and trillions, respectively, with one decimal place.

    Args:
    - num (int): The number to convert.

    Returns:
    - str: The converted, shortened number as a string.
    """
    if abs(num) < 1000:
        return str(num)
    elif abs(num) < 1000000:
        return f"{num / 1000:.1f}K"
    elif abs(num) < 1000000000:
        return f"{num / 1000000:.1f}M"
    elif abs(num) < 1000000000000:
        return f"{num / 1000000000:.1f}B"
    else:
        return f"{num / 1000000000000:.1f}T"


def model_summary(model: nn.Module, print_summary=False):
    "conver normal model to lightning model and print model summary"
    model = Model(model)
    summary = ModelSummary(model)
    if print_summary:
        print(summary)
    return {
        "summary": summary,
        "total_parameters": convert_int_to_shortened_string(summary.total_parameters),
        "trainable_parameters": convert_int_to_shortened_string(summary.trainable_parameters),
    }



# MIT License

# Copyright (c) 2024 Alexandre TL

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This file is taken from https://github.com/alxndrTL/mamba.py/tree/main
# curtosy of @AlexandreTL2 https://twitter.com/AlexandreTL2/

from __future__ import annotations

import torch
import torch.nn as nn
import torch.nn.functional as F
import math

from torch import Tensor


from dataclasses import dataclass

from ..modules.pscan import pscan

import math
from dataclasses import dataclass
from typing import Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from pscan import pscan

"""

This file closely follows the mamba_simple.py from the official Mamba implementation, and the mamba-minimal by @johnma2006.
The major differences are :
-the convolution is done with torch.nn.Conv1d
-the selective scan is done in PyTorch

A sequential version of the selective scan is also available for comparison.

- A Mamba model is composed of several layers, which are ResidualBlock.
- A ResidualBlock is composed of a MambaBlock, a normalization, and a residual connection : ResidualBlock(x) = mamba(norm(x)) + x
- This leaves us with the MambaBlock : its input x is (B, L, D) and its outputs y is also (B, L, D) (B=batch size, L=seq len, D=model dim).
First, we expand x into (B, L, 2*ED) (where E is usually 2) and split it into x and z, each (B, L, ED).
Then, we apply the short 1d conv to x, followed by an activation function (silu), then the SSM.
We then multiply it by silu(z).
See Figure 3 of the paper (page 8) for a visual representation of a MambaBlock.

"""


@dataclass
class MambaConfig:
    d_model: int  # D
    n_layers: int
    dt_rank: Union[int, str] = "auto"
    d_state: int = 16  # N in paper/comments
    expand_factor: int = 2  # E in paper/comments
    d_conv: int = 4

    dt_min: float = 0.001
    dt_max: float = 0.1
    dt_init: str = "random"  # "random" or "constant"
    dt_scale: float = 1.0
    dt_init_floor = 1e-4

    bias: bool = False
    conv_bias: bool = True

    pscan: bool = True  # use parallel scan mode or sequential mode when training

    def __post_init__(self):
        self.d_inner = self.expand_factor * self.d_model  # E*D = ED in comments

        if self.dt_rank == "auto":
            self.dt_rank = math.ceil(self.d_model / 16)


class Mamba(nn.Module):
    def __init__(self, config: MambaConfig):
        super().__init__()

        self.config = config

        self.layers = nn.ModuleList([ResidualBlock(config) for _ in range(config.n_layers)])
        # self.norm_f = RMSNorm(config.d_model)

    def forward(self, x):
        # x : (B, L, D)

        # y : (B, L, D)

        for layer in self.layers:
            x = layer(x)

        # x = self.norm_f(x)

        return x

    def step(self, x, caches):
        # x : (B, L, D)
        # caches : [cache(layer) for all layers], cache : (h, inputs)

        # y : (B, L, D)
        # caches : [cache(layer) for all layers], cache : (h, inputs)

        for i, layer in enumerate(self.layers):
            x, caches[i] = layer.step(x, caches[i])

        return x, caches


class ResidualBlock(nn.Module):
    def __init__(self, config: MambaConfig):
        super().__init__()

        self.mixer = MambaBlock(config)
        self.norm = RMSNorm(config.d_model)

    def forward(self, x):
        # x : (B, L, D)

        # output : (B, L, D)

        output = self.mixer(self.norm(x)) + x
        return output

    def step(self, x, cache):
        # x : (B, D)
        # cache : (h, inputs)
        # h : (B, ED, N)
        # inputs: (B, ED, d_conv-1)

        # output : (B, D)
        # cache : (h, inputs)

        output, cache = self.mixer.step(self.norm(x), cache)
        output = output + x
        return output, cache


class MambaBlock(nn.Module):
    def __init__(self, config: MambaConfig):
        super().__init__()

        self.config = config

        # projects block input from D to 2*ED (two branches)
        self.in_proj = nn.Linear(config.d_model, 2 * config.d_inner, bias=config.bias)

        self.conv1d = nn.Conv1d(
            in_channels=config.d_inner,
            out_channels=config.d_inner,
            kernel_size=config.d_conv,
            bias=config.conv_bias,
            groups=config.d_inner,
            padding=config.d_conv - 1,
        )

        # projects x to input-dependent Δ, B, C
        self.x_proj = nn.Linear(config.d_inner, config.dt_rank + 2 * config.d_state, bias=False)

        # projects Δ from dt_rank to d_inner
        self.dt_proj = nn.Linear(config.dt_rank, config.d_inner, bias=True)

        # dt initialization
        # dt weights
        dt_init_std = config.dt_rank**-0.5 * config.dt_scale
        if config.dt_init == "constant":
            nn.init.constant_(self.dt_proj.weight, dt_init_std)
        elif config.dt_init == "random":
            nn.init.uniform_(self.dt_proj.weight, -dt_init_std, dt_init_std)
        else:
            raise NotImplementedError

        # dt bias
        dt = torch.exp(
            torch.rand(config.d_inner) * (math.log(config.dt_max) - math.log(config.dt_min))
            + math.log(config.dt_min)
        ).clamp(min=config.dt_init_floor)
        inv_dt = dt + torch.log(
            -torch.expm1(-dt)
        )  # inverse of softplus: https://github.com/pytorch/pytorch/issues/72759
        with torch.no_grad():
            self.dt_proj.bias.copy_(inv_dt)
        # self.dt_proj.bias._no_reinit = True # initialization would set all Linear.bias to zero, need to mark this one as _no_reinit
        # todo : explain why removed

        # S4D real initialization
        A = torch.arange(1, config.d_state + 1, dtype=torch.float32).repeat(config.d_inner, 1)
        self.A_log = nn.Parameter(
            torch.log(A)
        )  # why store A in log ? to keep A < 0 (cf -torch.exp(...)) ? for gradient stability ?
        self.D = nn.Parameter(torch.ones(config.d_inner))

        # projects block output from ED back to D
        self.out_proj = nn.Linear(config.d_inner, config.d_model, bias=config.bias)

    def forward(self, x):
        # x : (B, L, D)

        # y : (B, L, D)

        _, L, _ = x.shape

        xz = self.in_proj(x)  # (B, L, 2*ED)
        x, z = xz.chunk(2, dim=-1)  # (B, L, ED), (B, L, ED)

        # x branch
        x = x.transpose(1, 2)  # (B, ED, L)
        x = self.conv1d(x)[:, :, :L]  # depthwise convolution over time, with a short filter
        x = x.transpose(1, 2)  # (B, L, ED)

        x = F.silu(x)
        y = self.ssm(x)

        # z branch
        z = F.silu(z)

        output = y * z
        output = self.out_proj(output)  # (B, L, D)

        return output

    def ssm(self, x):
        # x : (B, L, ED)

        # y : (B, L, ED)

        A = -torch.exp(self.A_log.float())  # (ED, N)
        D = self.D.float()
        # TODO remove .float()

        deltaBC = self.x_proj(x)  # (B, L, dt_rank+2*N)

        delta, B, C = torch.split(
            deltaBC,
            [self.config.dt_rank, self.config.d_state, self.config.d_state],
            dim=-1,
        )  # (B, L, dt_rank), (B, L, N), (B, L, N)
        delta = F.softplus(self.dt_proj(delta))  # (B, L, ED)

        if self.config.pscan:
            y = self.selective_scan(x, delta, A, B, C, D)
        else:
            y = self.selective_scan_seq(x, delta, A, B, C, D)

        return y

    def selective_scan(self, x, delta, A, B, C, D):
        # x : (B, L, ED)
        # Δ : (B, L, ED)
        # A : (ED, N)
        # B : (B, L, N)
        # C : (B, L, N)
        # D : (ED)

        # y : (B, L, ED)

        deltaA = torch.exp(delta.unsqueeze(-1) * A)  # (B, L, ED, N)
        deltaB = delta.unsqueeze(-1) * B.unsqueeze(2)  # (B, L, ED, N)

        BX = deltaB * (x.unsqueeze(-1))  # (B, L, ED, N)

        hs = pscan(deltaA, BX)

        y = (hs @ C.unsqueeze(-1)).squeeze(3)  # (B, L, ED, N) @ (B, L, N, 1) -> (B, L, ED, 1)

        y = y + D * x

        return y

    def selective_scan_seq(self, x, delta, A, B, C, D):
        # x : (B, L, ED)
        # Δ : (B, L, ED)
        # A : (ED, N)
        # B : (B, L, N)
        # C : (B, L, N)
        # D : (ED)

        # y : (B, L, ED)

        _, L, _ = x.shape

        deltaA = torch.exp(delta.unsqueeze(-1) * A)  # (B, L, ED, N)
        deltaB = delta.unsqueeze(-1) * B.unsqueeze(2)  # (B, L, ED, N)

        BX = deltaB * (x.unsqueeze(-1))  # (B, L, ED, N)

        h = torch.zeros(
            x.size(0), self.config.d_inner, self.config.d_state, device=deltaA.device
        )  # (B, ED, N)
        hs = []

        for t in range(0, L):
            h = deltaA[:, t] * h + BX[:, t]
            hs.append(h)

        hs = torch.stack(hs, dim=1)  # (B, L, ED, N)

        y = (hs @ C.unsqueeze(-1)).squeeze(3)  # (B, L, ED, N) @ (B, L, N, 1) -> (B, L, ED, 1)

        y = y + D * x

        return y

    # -------------------------- inference -------------------------- #
    """
    Concerning auto-regressive inference

    The cool part of using Mamba : inference is constant wrt to sequence length
    We just have to keep in cache, for each layer, two things :
    - the hidden state h (which is (B, ED, N)), as you typically would when doing inference with a RNN
    - the last d_conv-1 inputs of the layer, to be able to compute the 1D conv which is a convolution over the time dimension
      (d_conv is fixed so this doesn't incur a growing cache as we progress on generating the sequence)
      (and d_conv is usually very small, like 4, so we just have to "remember" the last 3 inputs)

    Concretely, these two quantities are put inside a cache tuple, and are named h and inputs respectively.
    h is (B, ED, N), and inputs is (B, ED, d_conv-1)
    The MambaBlock.step() receives this cache, and, along with outputing the output, alos outputs the updated cache for the next call.

    The cache object is initialized as follows : (None, torch.zeros()).
    When h is None, the selective scan function detects it and start with h=0.
    The torch.zeros() isn't a problem (it's same as just feeding the input, because the conv1d is padded)

    As we need one such cache variable per layer, we store a caches object, which is simply a list of cache object. (See mamba_lm.py)
    """

    def step(self, x, cache):
        # x : (B, D)
        # cache : (h, inputs)
        # h : (B, ED, N)
        # inputs : (B, ED, d_conv-1)

        # y : (B, D)
        # cache : (h, inputs)

        h, inputs = cache

        xz = self.in_proj(x)  # (B, 2*ED)
        x, z = xz.chunk(2, dim=1)  # (B, ED), (B, ED)

        # x branch
        x_cache = x.unsqueeze(2)
        x = self.conv1d(torch.cat([inputs, x_cache], dim=2))[
            :, :, self.config.d_conv - 1
        ]  # (B, ED)

        x = F.silu(x)
        y, h = self.ssm_step(x, h)

        # z branch
        z = F.silu(z)

        output = y * z
        output = self.out_proj(output)  # (B, D)

        # prepare cache for next call
        inputs = torch.cat([inputs[:, :, 1:], x_cache], dim=2)  # (B, ED, d_conv-1)
        cache = (h, inputs)

        return output, cache

    def ssm_step(self, x, h):
        # x : (B, ED)
        # h : (B, ED, N)

        # y : (B, ED)
        # h : (B, ED, N)

        A = -torch.exp(
            self.A_log.float()
        )  # (ED, N) # todo : ne pas le faire tout le temps, puisque c'est indépendant de la timestep
        D = self.D.float()
        # TODO remove .float()

        deltaBC = self.x_proj(x)  # (B, dt_rank+2*N)

        delta, B, C = torch.split(
            deltaBC,
            [self.config.dt_rank, self.config.d_state, self.config.d_state],
            dim=-1,
        )  # (B, dt_rank), (B, N), (B, N)
        delta = F.softplus(self.dt_proj(delta))  # (B, ED)

        deltaA = torch.exp(delta.unsqueeze(-1) * A)  # (B, ED, N)
        deltaB = delta.unsqueeze(-1) * B.unsqueeze(1)  # (B, ED, N)

        BX = deltaB * (x.unsqueeze(-1))  # (B, ED, N)

        if h is None:
            h = torch.zeros(
                x.size(0),
                self.config.d_inner,
                self.config.d_state,
                device=deltaA.device,
            )  # (B, ED, N)

        h = deltaA * h + BX  # (B, ED, N)

        y = (h @ C.unsqueeze(-1)).squeeze(2)  # (B, ED, N) @ (B, N, 1) -> (B, ED, 1)

        y = y + D * x

        # todo : pq h.squeeze(1) ??
        return y, h.squeeze(1)



from __future__ import annotations


from transformers import AutoTokenizer
from itertools import cycle
from datasets import load_from_disk, load_dataset

import torch
import torch.nn.functional as F
from torch.utils.data import IterableDataset, DataLoader
from pathlib import Path

import os
import requests
import random

PATH = Path("./data")
# "google/byt5-small"
# "NeelNanda/gpt-neox-tokenizer-digits"


def get_tokenizer(self, tokenizer: AutoTokenizer | str = None):
    self.tokenizer = tokenizer
    if tokenizer is None:
        tokenizer = "NeelNanda/gpt-neox-tokenizer-digits"
    if isinstance(tokenizer, str):
        self.tokenizer = AutoTokenizer.from_pretrained(
            tokenizer, cache_dir=self.cache_dir, use_fast=True
        )


class PreTokenizedDataset(IterableDataset):
    def __init__(
        self,
        dataset_name: str = "JeanKaddour/minipile",
        tokenizer: AutoTokenizer = None,
        split: str = "train",
        path: Path = PATH,
        microbatch_size: int = 32,
        min_length: int = 512,
        max_length: int = 2048,
        hf=False,
        cache_dir=None,
    ):
        self.tokenizer = tokenizer
        self.tokenizer.padding_side = "right"
        self.length = len(self.tokenizer)
        self.PAD = tokenizer.pad_token_id

        self.microbatch_size = microbatch_size
        self.vocab_size = len(tokenizer)
        self.min_length = min_length
        self.max_length = max_length + 1
        self.cache_dir = cache_dir
        self.dataset_name = dataset_name

        fpath = path
        if path == PATH:
            fpath = str(
                f"{self.dataset_name.replace('/','-')}--{self.tokenizer.name_or_path.replace('/','-')}"
            )
            fpath = str(path.joinpath(fpath).joinpath(split))
        if hf:
            self.ds = load_dataset(dataset_name)[split]
        else:
            self.ds = load_from_disk(fpath)
        self.toks_cycle = cycle(self.ds)

    def __iter__(self) -> torch.Tensor:
        while True:
            x = next(self.toks_cycle)["input_ids"]
            x = torch.tensor(x, dtype=torch.long)
            x = F.pad(x, (0, self.max_length - x.shape[0]), "constant", value=self.PAD)
            yield x[:-1][: self.max_length], x[1:][: self.max_length]


class TinyShakespeareDataset(IterableDataset):
    def __init__(
        self,
        tokenizer: AutoTokenizer = None,
        path: Path = PATH,
        min_length: int = 512,
        max_length: int = 512,
        cache_dir=None,
    ):
        assert tokenizer is not None, "must pass tokenizer"
        self.tokenizer = tokenizer
        self.tokenizer.padding_side = "right"
        self.length = len(self.tokenizer)
        self.PAD = tokenizer.pad_token_id

        self.vocab_size = len(tokenizer)
        self.min_length = min_length
        self.max_length = max_length
        self.cache_dir = cache_dir
        self.dataset_name = "tinyshakespeare"

        self.data_path = path.joinpath(self.dataset_name + ".txt")

        try:  # ugly ik
            with open(self.data_path) as f:
                self.data = torch.Tensor(tokenizer.encode(f.read())).long()
        except Exception:
            self.download_data()
            with open(self.data_path) as f:
                self.data = torch.Tensor(tokenizer.encode(f.read())).long()

        self.length = len(self.data)

    def __iter__(self) -> torch.Tensor:
        while True:
            idx = random.randint(0, (self.length - self.max_length - 1))
            x = self.data[idx : idx + self.max_length + 1]
            yield x[:-1][: self.max_length], x[1:][: self.max_length]

    def download_data(self):
        url = "https://raw.githubusercontent.com/karpathy/char-rnn/master/data/tinyshakespeare/input.txt"
        response = requests.get(url)
        if not os.path.exists(PATH):
            os.makedirs(PATH)
        if response.status_code == 200:
            with open(self.data_path, "w", encoding="utf-8") as file:
                file.write(response.text)
        else:
            raise Exception(f"Failed to download data. Status code: {response.status_code}")


if __name__ == "__main__":
    tokenizer = AutoTokenizer.from_pretrained("microsoft/phi-2")
    dataset = PreTokenizedDataset(
        dataset_name="roneneldan/TinyStories", tokenizer=tokenizer, cache_dir="hf_cache"
    )
    dataloder = DataLoader(dataset, batch_size=2)

    for data, target in dataloder:
        print(data.shape, target.shape)
        exit(0)
        # print(tokenizer.decode(data.tolist()))



import jax
import jax.numpy as jnp
import einops
from jaxtyping import Array


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0) -> tuple[Array, Array]:
    freqs = 1.0 / (10000 ** (jnp.arange(0, dim, 2) / dim))
    freqs = jnp.einsum("i , j -> i j", jnp.arange(end), freqs)
    freqs_cos = jnp.cos(freqs)  # real part
    freqs_sin = jnp.sin(freqs)  # imaginary part
    return freqs_cos, freqs_sin


def rotate_half(x) -> tuple[Array, Array]:
    x1 = x[:, ::2]
    x2 = x[:, 1::2]
    x = jnp.stack((-x2, x1), axis=-1)
    return einops.rearrange(x, "... d j -> ... (d j)")


def apply_rope(q: Array, k, freq_cis) -> tuple[Array, Array]:
    cos, sin = (einops.repeat(t, "b n -> b (n j)", j=2) for t in freq_cis)
    print(q.shape, k.shape, cos.shape, sin.shape)
    sin_q = sin[-q.shape[0] :, :]
    cos_q = cos[-q.shape[0] :, :]
    print(q.shape, k.shape, cos.shape, sin.shape)

    q = (q * cos_q) + (rotate_half(q) * sin_q)
    k = (k * cos) + (rotate_half(k) * sin)

    return q, k


if __name__ == "__main__":
    B, T, nh, C = 3, 5, 4, 16

    key = jax.random.PRNGKey(0)
    k = jax.random.normal(key, (B, T, C)).reshape(B, T, nh, C // nh)
    q = jax.random.normal(key, (B, T, C)).reshape(B, T, nh, C // nh)

    freq_cis = precompute_freqs_cis(C // nh, T)

    xq_out, xk_out = apply_rope(k, q, freq_cis)



import lightning as L
from huggingface_hub import PyTorchModelHubMixin
from transformers import AutoModel
import ohara.models.phi as phi
import torch


class Transformer(L.LightningModule, PyTorchModelHubMixin):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, *args, **kwargs):
        return self.model(**args, **kwargs)

    @classmethod
    def load_phi(cls, model_name="microsoft/phi-2"):
        cfg = phi.PhiConfig()
        model = phi.Phi(cfg)
        hf_model = AutoModel.from_pretrained(model_name)
        hf_model
        mask = torch.full((1, 1, cfg.seq_len, cfg.seq_len), float("-inf"))

        mode



from transformers import AutoTokenizer
import jax.numpy as np

tokenizer = AutoTokenizer.from_pretrained("NeelNanda/gpt-neox-tokenizer-digits", use_fast=True)
tokenizer.padding_side = "right"


vocab_size = len(tokenizer)
x_d = 768
x_n_min = 512
x_n_max = 2048
m_n = 64
layers = 1


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
    freqs = 1.0 / (10000 ** (jnp.arange(0, dim, 2) / dim))
    freqs = jnp.einsum("i , j -> i j", jnp.arange(end), freqs)
    freqs_cos = np.cos(freqs)  # real part
    freqs_sin = np.sin(freqs)  # imaginary part
    return freqs_cos, freqs_sin


def rotate_half(x):
    x1 = x[:, ::2]
    x2 = x[:, 1::2]

    x = jnp.stack((-x2, x1), axis=-1)

    return rearrange(x, "... d j -> ... (d j)")


def apply_rope(q, k, freq_cis):
    cos, sin = (repeat(t, "b n -> b (n j)", j=2) for t in freq_cis)
    print(q.shape, k.shape, cos.shape, sin.shape)
    sin_q = sin[-q.shape[0] :, :]
    cos_q = cos[-q.shape[0] :, :]
    print(q.shape, k.shape, cos.shape, sin.shape)

    q = (q * cos_q) + (rotate_half(q) * sin_q)
    k = (k * cos) + (rotate_half(k) * sin)

    return q, k


B, T, C = 3, 5, 64
nh = 4
hs = C // nh
q = jax.random.uniform(random.PRNGKey(0), shape=(T, C))
k = jax.random.uniform(random.PRNGKey(0), shape=(T, C))

q = q.reshape(T, nh, hs)
k = k.reshape(T, nh, hs)

freq_cis = precompute_freqs_cis(dim=hs, end=T)
q1, k1 = apply_rope(q, k, freq_cis)

# exit(0)


import torch


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
    t = torch.arange(end, device=freqs.device)  # type: ignore
    freqs = torch.outer(t, freqs).float()  # type: ignore
    freqs_cos = torch.cos(freqs)  # real part
    freqs_sin = torch.sin(freqs)  # imaginary part
    return freqs_cos, freqs_sin


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
    ndim = x.ndim
    assert 0 <= 1 < ndim
    assert freqs_cis.shape == (x.shape[1], x.shape[-1])
    shape = [d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)]
    return freqs_cis.view(shape)


def apply_rotary_emb(
    xq: torch.Tensor, xk: torch.Tensor, freqs_cos: torch.Tensor, freqs_sin: torch.Tensor
) -> tuple[torch.Tensor, torch.Tensor]:
    # reshape xq and xk to match the complex representation
    xq_r, xq_i = xq.float().reshape(xq.shape[:-1] + (-1, 2)).unbind(-1)
    xk_r, xk_i = xk.float().reshape(xk.shape[:-1] + (-1, 2)).unbind(-1)

    # reshape freqs_cos and freqs_sin for broadcasting
    freqs_cos = reshape_for_broadcast(freqs_cos, xq_r)
    freqs_sin = reshape_for_broadcast(freqs_sin, xq_r)

    # apply rotation using real numbers
    xq_out_r = xq_r * freqs_cos - xq_i * freqs_sin
    xq_out_i = xq_r * freqs_sin + xq_i * freqs_cos
    xk_out_r = xk_r * freqs_cos - xk_i * freqs_sin
    xk_out_i = xk_r * freqs_sin + xk_i * freqs_cos

    # flatten last two dimensions
    xq_out = torch.stack([xq_out_r, xq_out_i], dim=-1).flatten(3)
    xk_out = torch.stack([xk_out_r, xk_out_i], dim=-1).flatten(3)

    return xq_out.type_as(xq), xk_out.type_as(xk)


q, k = torch.tensor(q.tolist()), torch.tensor(k.tolist())
T, C = q.shape
nh = 4
hs = C // nh
q = q.reshape(1, T, nh, hs)
k = k.reshape(1, T, nh, hs)

freq_cis = precompute_freqs_cis(hs, 5)
q1, k1 = apply_rotary_emb(q, k, freq_cis[0], freq_cis[1])
print(q1.shape)



from transformers import AutoTokenizer
import optax
import wandb
import sys
from tqdm import tqdm
from itertools import cycle
from datasets import load_dataset
import jax.experimental.mesh_utils as mesh_utils
import jax.sharding as sharding
from ramen import *

# wandb.login()

so = open("data.log", "w", 10)
sys.stdout.echo = so
sys.stderr.echo = so


ds = load_dataset("JeanKaddour/minipile", split="train")
ds_val = load_dataset("JeanKaddour/minipile", split="validation")

tokenizer = AutoTokenizer.from_pretrained("NeelNanda/gpt-neox-tokenizer-digits", use_fast=True)
tokenizer.padding_side = "right"

PAD = tokenizer.pad_token_id

microbatch_size = 32
accum_batch_size = microbatch_size * 4
vocab_size = len(tokenizer)
x_d = 768
x_n_min = 512
x_n_max = 2048
m_n = 64
layers = 14
dtype = jnp.bfloat16


def tokenize_fn(x):
    return tokenizer(x["text"], max_length=x_n_max, truncation=True)


def filter_fn(x):
    return len(x["input_ids"]) >= x_n_min


toks = (
    ds.map(tokenize_fn, batched=True, remove_columns=["text"], num_proc=128)
    .shuffle(seed=42)
    .filter(filter_fn, num_proc=128)
)
toks_val = ds_val.map(tokenize_fn, batched=True, remove_columns=["text"], num_proc=128).filter(
    filter_fn, num_proc=128
)


@eqx.filter_jit
def loss_fn_val(model, batch):
    seq = jax.vmap(lambda x: jnp.concatenate([jnp.array([PAD]), x]))(batch[:, :-1])

    logits = jax.vmap(model)(seq)

    losses = optax.softmax_cross_entropy_with_integer_labels(logits, batch)
    mask = jnp.where(batch != PAD, 1, 0)
    losses = losses * mask
    return losses.sum() / mask.sum()


loss_fn = eqx.filter_value_and_grad(loss_fn_val)

model = RAMENModel(vocab_size, layers, x_d, m_n, key=random.PRNGKey(0))

B, T, C = (3, 5, x_d)
seq = jax.random.uniform(random.PRNGKey(0), shape=(B, T), minval=0, maxval=vocab_size)
logits = jax.vmap(model)(seq)
exit(0)
if dtype != jnp.float32:
    model = jax.tree_util.tree_map(lambda x: x.astype(dtype) if eqx.is_array(x) else x, model)


toks_cycle = cycle(toks)


def get_microbatch():
    microbatch = [next(toks_cycle)["input_ids"] for i in range(microbatch_size)]
    microbatch = [jnp.array(x) for x in microbatch]
    max_len = max([x.shape[0] for x in microbatch])
    microbatch = [jnp.pad(x, (0, max_len - x.shape[0]), constant_values=PAD) for x in microbatch]
    microbatch = jnp.stack(microbatch)
    return microbatch


num_microbatches = accum_batch_size // microbatch_size
num_epochs = 1
num_batches = num_epochs * len(toks) // accum_batch_size

schedule = optax.cosine_onecycle_schedule(num_batches, 2e-4)
optimizer = optax.chain(optax.clip_by_global_norm(1.0), optax.lion(schedule))
optim = optax.MultiSteps(optimizer, every_k_schedule=num_microbatches)

opt_state = optim.init(eqx.filter(model, eqx.is_array))


def Updater(optimizer):
    @eqx.filter_jit
    def train_step(model, opt_state, batch):
        loss, grads = loss_fn(model, batch)
        model_arrays = eqx.filter(model, eqx.is_array)
        updates, opt_state = optimizer.update(grads, opt_state, model_arrays)
        model_arrays = eqx.apply_updates(model_arrays, updates)

        model = eqx.combine(model_arrays, model)

        return loss, model, opt_state

    return train_step


params_count = sum(x.size for x in jax.tree_util.tree_leaves(eqx.filter(model, eqx.is_array)))
print(f"{params_count / 1000000}M Parameters")

train_step = Updater(optim)


def validate(model, toks):
    acum_loss = 0.0

    idx = 0
    count = len(toks_val) // microbatch_size

    for i in range(count):
        microbatch = [toks_val[idx + i]["input_ids"] for i in range(microbatch_size)]
        microbatch = [jnp.array(x) for x in microbatch]
        max_len = max([x.shape[0] for x in microbatch])
        microbatch = [
            jnp.pad(x, (0, max_len - x.shape[0]), constant_values=PAD) for x in microbatch
        ]
        microbatch = jnp.stack(microbatch)
        microbatch = jax.device_put(microbatch, shard)
        idx += microbatch_size

        acum_loss += loss_fn_val(model, microbatch)

    acum_loss /= count
    return acum_loss


num_devices = len(jax.devices())
devices = mesh_utils.create_device_mesh((num_devices, 1))
shard = sharding.PositionalSharding(devices)


# run = wandb.init(project='memnet-small-1epoch', entity='jpritsk')

for epoch in range(num_epochs):
    for batch in tqdm(range(num_batches)):
        accum_loss = 0.0

        for _i in range(num_microbatches):
            microbatch = get_microbatch()
            microbatch = jax.device_put(microbatch, shard)
            loss, model, opt_state = train_step(model, opt_state, microbatch)
            accum_loss += loss
            assert not jnp.isnan(accum_loss.item())

        accum_loss /= num_microbatches

        if (batch + 1) % 20 == 0:
            wandb.log({"train_loss": accum_loss.item()}, step=batch + 1)

        if (batch + 1) % 100 == 0 or batch == num_batches - 1:
            val_loss = validate(model, toks_val)
            # wandb.log({'val_loss': val_loss.item()}, step=batch + 1)
            print(f"\nStep: {batch + 1}/{num_batches}, loss: {val_loss.item()}")

    eqx.tree_serialise_leaves(f"model{epoch}.eqx", model)
    wandb.save(self_path)
    wandb.save(f"model{epoch}.eqx")

# wandb.finish()



from __future__ import annotations

import os

# Load model directly
from huggingface_hub import snapshot_download

model_name = "microsoft/phi-2"


def get_model_path(model_name: str) -> tuple[str, str | None]:
    CACHE_DIR = os.path.expanduser("~/.cache/huggingface/hub/")
    model_name = os.path.join("models--" + model_name.replace("/", "--"), "snapshots")
    path = os.path.join(CACHE_DIR, model_name)
    if not os.path.exists(path):
        return None, "Not exits"
    return os.path.join(path, os.listdir(path)[0]), None


def download_hf_model(model_name: str) -> str:
    """
    download model from huggingface and return path of saved model
    """

    return snapshot_download(model_name)


if __name__ == "__main__":
    print(download_hf_model(model_name))



import torch
import torch.nn as nn

from transformers import AutoTokenizer


def smapler(logits, *args, **kwargs): ...


def generate(model: nn.Module, tokenizer: AutoTokenizer, max_tokens: int = 64, *args, **kwargs):
    """
    Generate text from a model
    """
    model.eval()

    for _ in range(max_tokens):
        input_ids = torch.tensor(tokenizer.encode("")).unsqueeze(0)
        with torch.no_grad():
            outputs = model.generate(input_ids, **kwargs)
            tokenizer.decode(outputs[0].tolist())



from re import DEBUG
import torch
import torch.nn as nn

from ohara.modules.mlp import MLP_MAP


# This might not me most efficient implementation of MOE
# but it is easy to understand
# TODO: Write a more efficient implementation and more types or moe

DEBUG = False

class MoE(nn.Module):
    def __init__(
        self,
        dim: int,
        hidden_dim: int | None = None,
        num_experts: int = 4,
        num_experts_per_tok: int = 2,
        mlp: str = "swiglu",
    ):
        super().__init__()
        self.num_experts = num_experts
        self.num_experts_per_tok = num_experts_per_tok

        mlp_block = MLP_MAP[mlp]  # SwiGLU is default

        self.experts = nn.ModuleList([mlp_block(dim, hidden_dim) for i in range(num_experts)])
        self.gate = nn.Linear(dim, num_experts, bias=False)

    def forward(self, x: torch.Tensor):
        batch_size, seq_len, dim = x.shape  # (batch_size, seq_len, dim)

        # (batch_size , seq_len, dim) -> (batch_size * seq_len, dim)
        x = x.view(batch_size * seq_len, dim)

        # (batch_size * seq_len, dim) -> (batch_size * seq_len, num_experts)
        scores = self.gate(x)

        # expert_weights -> (batch_size * seq_len, num_experts_per_tok)
        # expert_indices -> (batch_size * seq_len, num_experts_per_tok)
        expert_weights, expert_indices = torch.topk(scores, self.num_experts_per_tok, dim=-1)

        if DEBUG:
            lst = [0]*4
            for _idx in expert_indices.reshape(-1):
                lst[_idx]+=1
            for idx,item in enumerate(lst):
                print(f"{idx=} {item=}")
            print("-------------")

        # -> (batch_size * seq_len, num_experts_per_tok)
        expert_weights = expert_weights.softmax(dim=-1)

        #  -> (batch_size * seq_len * num_experts_per_tok ) 1D
        flat_expert_indices = expert_indices.view(-1)

        # (batch_size * seq_len, dim) -> (batch_size * seq_len * num_experts_per_tok, dim)
        # create copied of inputs for each expert
        x = x.repeat_interleave(self.num_experts_per_tok, dim=0)

        # (total_tokens,dim)
        output = torch.empty_like(x, dtype=x.dtype, device=x.device)

        for idx, expert in enumerate(self.experts):
            # filtered_x - selected toks that to be sent to nth expert
            filtered_x = x[flat_expert_indices == idx]
            output[flat_expert_indices == idx] = expert(filtered_x)

        # ->B,T,num_experts_per_tok,dim
        output = output.view(*expert_weights.shape, -1)
        expert_weights = expert_weights.unsqueeze(-1)

        output = output * expert_weights

        # sum up experts outputs
        # batch_size * seq_len, num_experts_per_tok, dim -> batch_size * seq_len, dim
        output = output.sum(dim=1)

        return output.view(batch_size, seq_len, dim)  #  batch_size, seq_len, dim


if __name__ == "__main__":
    B, T, C = 8, 16, 32 

    with torch.device("cuda"):
        import torch.optim as optim
        torch.set_float32_matmul_precision('high')
        model = MoE(C, C)
        model = torch.compile(model)
        optimizer = optim.AdamW(model.parameters())

        x = torch.randn(B, T, C)
        y = torch.randn(B, T, C)

        iters = 1000

        DEBUG = True
        model(x)
        DEBUG = False

        for _ in range(iters):
            p = model(x)
            loss = torch.nn.functional.mse_loss(p,y)
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
            # print(f"{loss.item()=}")

        DEBUG = True
        model(x)
        DEBUG = False


        print(y.shape)



import jax
import jax.numpy as jnp
import equinox as eqx

from jaxtyping import Array


class RMSNorm(eqx.Module):
    eps: float
    weight: Array

    def __init__(self, dim, eps=1e-6, *, key):
        self.eps = eps
        self.weight = jnp.ones(dim)

    def __call__(self, x: Array):
        x = x * (1 / jnp.sqrt(jnp.power(x, 2).mean(-1, keepdims=True) + self.eps))
        return self.weight * x


if __name__ == "__main__":
    key = jax.random.PRNGKey(0)
    B, T, C = 3, 5, 256
    model = RMSNorm(dim=C, key=key)
    x = jax.random.normal(key, (B, T, C))

    out = jax.vmap(model)(x)

    print(out.shape)



from __future__ import annotations

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

import math

from ohara.embedings_pos.rotatry import apply_rope

TensorTuple = tuple[Tensor, ...]


class CasualAttention(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        attn_dropout: float = 0.0,
        res_dropout: float = 0.0,
        idx: int | None = None,
    ) -> None:
        super().__init__()
        self.num_heads: int = num_heads
        self.head_dim: int = d_model // num_heads
        self.idx = idx

        self.key = nn.Linear(d_model, d_model)
        self.query = nn.Linear(d_model, d_model)
        self.value = nn.Linear(d_model, d_model)
        self.proj = nn.Linear(d_model, d_model)

        self.attn_dropout = nn.Dropout(attn_dropout)
        self.res_dropout = nn.Dropout(res_dropout)

        self.flash_attn: bool = hasattr(torch.nn.functional, "scaled_dot_product_attention")

    def forward(
        self,
        x: torch.Tensor,
        mask: torch.Tensor = None,
        freqs_cis: TensorTuple | None = None,
        verbose: bool = False,
        **kwargs: dict,
    ) -> torch.Tensor:
        batch, seq_len, d_model = x.shape

        k: torch.Tensor  # type hint for lsp
        q: torch.Tensor  # ignore
        v: torch.Tensor

        k = self.key(x)
        q = self.query(x)
        v = self.value(x)

        k = k.view(
            batch, seq_len, self.num_heads, self.head_dim
        )  # shape = (B, seq_len, num_heads, head_dim)
        q = q.view(batch, seq_len, self.num_heads, self.head_dim)
        v = v.view(batch, seq_len, self.num_heads, self.head_dim)

        q, k = apply_rope(q, k, freqs_cis)

        # xk = repeat_kv(xk, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)
        # xv = repeat_kv(xv, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)

        k = k.transpose(1, 2)  # shape = (B, num_heads, seq_len, head_dim)
        q = q.transpose(1, 2)
        v = v.transpose(1, 2)

        if self.flash_attn and not verbose:
            output = torch.nn.functional.scaled_dot_product_attention(
                q,
                k,
                v,  # order impotent
                attn_mask=None,
                dropout_p=self.attn_dropout.p if self.training else 0.0,
                is_causal=True,
            )
        else:
            attn_mtx = torch.matmul(q, k.transpose(2, 3)) / math.sqrt(self.head_dim)
            attn_mtx = attn_mtx + mask[:, :, :seq_len, :seq_len]
            attn_mtx = F.softmax(attn_mtx.float(), dim=-1).type_as(k)
            attn_mtx_dropout = self.attn_dropout(attn_mtx)

            output = torch.matmul(attn_mtx_dropout, v)  # (batch, n_head, seq_len, head_dim)

        # restore time as batch dimension and concat heads
        output = output.transpose(1, 2).contiguous().view(batch, seq_len, d_model)

        # final projection into the residual stream
        output = self.proj(output)
        output = self.res_dropout(output)

        if verbose:
            return output, {"idx": self.idx, "attn_mtx": attn_mtx}
        return output


class Attention(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        attn_dropout: float = 0.0,
        res_dropout: float = 0.0,
        idx: int | None = None,
    ) -> None:
        super().__init__()
        self.num_heads: int = num_heads
        self.head_dim: int = d_model // num_heads
        self.idx = idx

        self.key = nn.Linear(d_model, d_model)
        self.query = nn.Linear(d_model, d_model)
        self.value = nn.Linear(d_model, d_model)
        self.proj = nn.Linear(d_model, d_model)

        self.attn_dropout = nn.Dropout(attn_dropout)
        self.res_dropout = nn.Dropout(res_dropout)

        self.flash_attn: bool = hasattr(torch.nn.functional, "scaled_dot_product_attention")

    def forward(
        self,
        x: torch.Tensor,
        mask: torch.Tensor = None,
        freqs_cis: TensorTuple | None = None,
        verbose: bool = False,
        **kwargs: dict,
    ) -> torch.Tensor:
        batch, seq_len, d_model = x.shape

        k: torch.Tensor  # type hint for lsp
        q: torch.Tensor  # ignore
        v: torch.Tensor

        k = self.key(x)
        q = self.query(x)
        v = self.value(x)

        k = k.view(
            batch, seq_len, self.num_heads, self.head_dim
        )  # shape = (B, seq_len, num_heads, head_dim)
        q = q.view(batch, seq_len, self.num_heads, self.head_dim)
        v = v.view(batch, seq_len, self.num_heads, self.head_dim)

        q, k = apply_rope(q, k, freqs_cis)

        # xk = repeat_kv(xk, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)
        # xv = repeat_kv(xv, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)

        k = k.transpose(1, 2)  # shape = (B, num_heads, seq_len, head_dim)
        q = q.transpose(1, 2)
        v = v.transpose(1, 2)

        if mask is None and self.flash_attn:
            ...

        if self.flash_attn and not verbose:
            output = torch.nn.functional.scaled_dot_product_attention(
                q,
                k,
                v,  # order impotent
                attn_mask=None,
                dropout_p=self.attn_dropout.p if self.training else 0.0,
                is_causal=True,
            )
        else:
            attn_mtx = torch.matmul(q, k.transpose(2, 3)) / math.sqrt(self.head_dim)
            attn_mtx = attn_mtx + mask[:, :, :seq_len, :seq_len]
            attn_mtx = F.softmax(attn_mtx.float(), dim=-1).type_as(k)
            attn_mtx_dropout = self.attn_dropout(attn_mtx)

            output = torch.matmul(attn_mtx_dropout, v)  # (batch, n_head, seq_len, head_dim)

        # restore time as batch dimension and concat heads
        output = output.transpose(1, 2).contiguous().view(batch, seq_len, d_model)

        # final projection into the residual stream
        output = self.proj(output)
        output = self.res_dropout(output)

        if verbose:
            return output, {"idx": self.idx, "attn_mtx": attn_mtx}
        return output


class CrossAttention(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        attn_dropout: float = 0.0,
        res_dropout: float = 0.0,
        idx: int | None = None,
    ) -> None:
        super().__init__()
        self.num_heads: int = num_heads
        self.head_dim: int = d_model // num_heads
        self.idx = idx
        self.query = nn.Linear(d_model, d_model)

        self.attn_dropout = nn.Dropout(attn_dropout)
        self.res_dropout = nn.Dropout(res_dropout)

        self.flash_attn: bool = hasattr(torch.nn.functional, "scaled_dot_product_attention")

    def forward(
        self,
        x: torch.Tensor,
        query: torch.Tensor,
        value: torch.Tensor,
        mask: torch.Tensor = None,
        freqs_cis: TensorTuple | None = None,
        is_casual: bool = True,
        verbose: bool = False,
        **kwargs: dict,
    ) -> torch.Tensor:
        batch, seq_len, d_model = x.shape

        k: torch.Tensor = self.key(x)
        k = k.view(
            batch, seq_len, self.num_heads, self.head_dim
        )  # shape = (B, seq_len, num_heads, head_dim)
        q = query.view(batch, seq_len, self.num_heads, self.head_dim)
        v = value.view(batch, seq_len, self.num_heads, self.head_dim)

        q, k = apply_rope(q, k, freqs_cis)

        # xk = repeat_kv(xk, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)
        # xv = repeat_kv(xv, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)

        k = k.transpose(1, 2)  # shape = (B, num_heads, seq_len, head_dim)
        q = q.transpose(1, 2)
        v = v.transpose(1, 2)

        # restore time as batch dimension and concat heads
        output = output.transpose(1, 2).contiguous().view(batch, seq_len, d_model)

        # final projection into the residual stream
        output = self.proj(output)
        output = self.res_dropout(output)

        if verbose:
            return output, {"idx": self.idx, "attn_mtx": attn_mtx}
        return output



from __future__ import annotations

import torch
import torch.nn as nn
from torch import Tensor

import math


class RotatryEmbedding(nn.Module):
    def __init__(self, dim: int, max_seq_len: int = 2048):
        super().__init__()
        self.dim = dim
        self.max_seq_len = max_seq_len

        cis = RotatryEmbedding.precompute_freqs_cis(dim, max_seq_len)
        self.register_buffer("cos", cis[0])
        self.register_buffer("sin", cis[1])

    @staticmethod
    def rotate_half(x: Tensor):
        """Rotates half the hidden dims of the input."""
        x1 = x[..., : x.shape[-1] // 2]
        x2 = x[..., x.shape[-1] // 2 :]
        return torch.cat((-x2, x1), dim=-1)

    @staticmethod
    def apply_rotary_pos_emb(
        x: Tensor,
        cos: Tensor,
        sin: Tensor,
        position_ids: int | None = None,
        unsqueeze_dim: int = 1,
    ):
        cos = cos[position_ids].unsqueeze(unsqueeze_dim)
        sin = sin[position_ids].unsqueeze(unsqueeze_dim)
        x_embed = (x * cos) + (RotatryEmbedding.rotate_half(x) * sin)
        return x_embed

    def forward(self, x, position_ids=None):
        return self.apply_rotary_pos_emb(x, self.cos, self.sin, position_ids=position_ids)

    @staticmethod
    def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
        freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
        # [: (dim // 2)] for odd number truncation
        # torch.arange(0, dim, 2) -> 2(i-1)//d while i= 1,2,..,(d//2)

        t = torch.arange(end, device=freqs.device)
        freqs = torch.outer(t, freqs).float()  # gives diffrent angle vector

        # e^it = cos(t) + i sin(t)
        freqs_cos = torch.cos(freqs)  # real
        freqs_sin = torch.sin(freqs)  # imaginary
        return freqs_cos, freqs_sin


# rotary embedding
def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
    # [: (dim // 2)] for odd number truncation
    # torch.arange(0, dim, 2) -> 2(i-1)//d while i= 1,2,..,(d//2)

    t = torch.arange(end, device=freqs.device)
    freqs = torch.outer(t, freqs).float()  # gives diffrent angle vector

    # e^it = cos(t) + i sin(t)
    freqs_cos = torch.cos(freqs)  # real
    freqs_sin = torch.sin(freqs)  # imaginary
    return freqs_cos, freqs_sin


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
    ndim = x.dim()
    assert 1 < ndim
    assert freqs_cis.shape == (
        x.shape[1],
        x.shape[-1],
    ), f"{freqs_cis.shape=}, {(x.shape[1], x.shape[-1])=}"

    # keep 2nd (T) and last(freq) dim same else make dim 1 for freq_cis
    shape = [d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)]
    # print(shape)
    return freqs_cis.view(shape)


def apply_rope(q,k, cis):
    # Idea suppose vector v = [x,y,x1,y1,...] # v.shape = dim
    # convert vetor into complex num # ie two vec one real, one imagery
    # [x,y,x1,y1,...] -> x+iy, x1+iy1
    # Multiplying by complex num == roatate vector
    # => (x + iy) * (cos + isin) -> x'+iy'
    # restack
    # x'+iy' -> [x',y',x1',y1'...]
    # you roated vector in chunks of two lfg!!!
    _, seq_len, _, _ = q.shape

    freqs_cos, freqs_sin = cis
    freqs_cos, freqs_sin = freqs_cos[:seq_len], freqs_sin[:seq_len]

    #  rehsape a shape (...,n )-> (..., n//2,2)
    q_cis = q.float().reshape(
        q.shape[:-1] + (-1, 2)
    )  # (B,T,nhead,C) -> (B,T,nhead,Cc,2) # Cc = C//2
    k_cis = k.float().reshape(k.shape[:-1] + (-1, 2))  # (B,T,nhead,C) -> (B,T,nhead,Cc,2)

    xq_r, xq_i = q_cis.unbind(-1)  # (B,T,nhead,Cc,2) -> ((B,T,Cc), (B,T,Cc)) split into two tuple
    xk_r, xk_i = k_cis.unbind(-1)  # (B,T,nhead,Cc,2) -> ((B,T,Cc), (B,T,Cc))

    freqs_cos = reshape_for_broadcast(freqs_cos, xq_r)  # freqs.shape = (1,T,1,Cc)
    freqs_sin = reshape_for_broadcast(freqs_sin, xq_r)

    # e+if = (a+ib) * (c+di) = (ac-bd) + i (ad+bc)
    # a = xq_r , b = xq_i
    # c = fcos , d = fsin
    # ...
    # e = (ac-bd) = xq_r * freqs_cos - xq_i * freqs_sin
    # f = (c+di)  = xq_r * freqs_sin + xq_i * freqs_cos

    xq_out_r = xq_r * freqs_cos - xq_i * freqs_sin  # (ac-bd)   # shape =  # (B,T,nhead,Cc)
    xq_out_i = xq_r * freqs_sin + xq_i * freqs_cos  # (ad+bc) * i
    xk_out_r = xk_r * freqs_cos - xk_i * freqs_sin  # (ac-bd)
    xk_out_i = xk_r * freqs_sin + xk_i * freqs_cos  # (ad+bc) * i

    # now we stack r,i -> [r,i,r2,i2]
    xq_out = torch.stack([xq_out_r, xq_out_i], dim=-1)  # (B,T,nhead,Cc,2)
    xk_out = torch.stack([xk_out_r, xk_out_i], dim=-1)  # (B,T,nhead,Cc,2)

    # flatten last two dimensions
    xq_out = xq_out.flatten(3)  # (B,T,nhead,C)
    xk_out = xk_out.flatten(3)  # (B,T,nhead,C)

    return xq_out.type_as(q), xk_out.type_as(q)


class RoPE(nn.Module):
    """Implements the rotary positional encoding.

    The traditional implementation rotates consecutive pairs of elements in the
    feature dimension while the default implementation rotates pairs with
    stride half the feature dimensions for efficiency.

    For more details see `RoFormer: Enhanced Transformer with Rotary Position
    Embedding <https://arxiv.org/abs/2104.09864>`_.

    Args:
        dims (int): The feature dimensions to be rotated. If the input feature
            is larger than dims then the rest is left unchanged.
        traditional (bool, optional): If set to True choose the traditional
            implementation which is slightly less efficient. Default: ``False``.
        base (float, optional): The base used to compute angular frequency for
            each dimension in the positional encodings. Default: ``10000``.
        scale (float, optional): The scale used to scale the positions. Default: ``1.0``.
    """

    def __init__(
        self,
        dims: int,
        traditional: bool = False,
        base: float = 10000,
        scale: float = 1.0,
    ):
        super().__init__()

        self.dims = dims
        self.traditional = traditional
        self.base = base
        self.scale = scale

    def _extra_repr(self):
        return f"{self.dims}, traditional={self.traditional}"

    def _compute_rope(self, costheta, sintheta, x):
        x1 = x[..., : self.dims // 2]
        x2 = x[..., self.dims // 2 : self.dims]
        rx1 = x1 * costheta - x2 * sintheta
        rx2 = x1 * sintheta + x2 * costheta

        if self.dims < x.shape[-1]:
            rx = torch.concatenate([rx1, rx2, x[..., self.dims :]], axis=-1)
        else:
            rx = torch.concatenate([rx1, rx2], axis=-1)

        return rx

    def _compute_traditional_rope(self, costheta, sintheta, x):
        x1 = x[..., ::2]
        x2 = x[..., 1::2]
        rx1 = x1 * costheta - x2 * sintheta
        rx2 = x1 * sintheta + x2 * costheta

        if self.dims < x.shape[-1]:
            raise NotImplementedError("RoPE doesn't implement partial traditional application")

        rx = torch.cat([rx1[..., None], rx2[..., None]], axis=-1)

        return rx

    def forward(self, x: Tensor, offset: int = 0):
        shape = x.shape
        x = x.reshape(-1, shape[-2], shape[-1])
        N = x.shape[1] + offset
        costheta, sintheta = RoPE.create_cos_sin_theta(
            N, self.dims, offset=offset, base=self.base, scale=self.scale, dtype=x.dtype
        )

        rope = self._compute_traditional_rope if self.traditional else self._compute_rope
        rx = rope(costheta, sintheta, x)

        return torch.reshape(rx, shape)

    @staticmethod
    def create_cos_sin_theta(
        N: int,
        D: int,
        offset: int = 0,
        base: float = 10000,
        scale: float = 1.0,
        dtype=torch.float32,
    ):
        D = D // 2
        positions = torch.arange(offset, N, dtype=dtype) * scale
        freqs = torch.exp(-torch.arange(0.0, D, dtype=dtype) * (math.log(base) / D))
        theta = torch.reshape(positions, (-1, 1)) * torch.reshape(freqs, (1, -1))
        return torch.cos(theta), torch.sin(theta)



import jax
import jax.numpy as np
import equinox as eqx

from jaxtyping import Array
from rope import precompute_freqs_cis, apply_rope
from ffn import SwiGLU
from norm import RMSNorm


class Attention(eqx.Module):
    dim: int
    num_heads: int
    head_dim: int
    seq_len: int
    k: eqx.nn.Linear
    q: eqx.nn.Linear
    v: eqx.nn.Linear
    proj: eqx.nn.Linear
    freqs_cis: Array

    def __init__(self, dim, num_heads, seq_len, *, key):
        super().__init__()
        key_k, key_q, key_v, key_proj = jax.random.split(key, 4)
        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.seq_len = seq_len
        self.k = eqx.nn.Linear(dim, dim, key=key_k)
        self.q = eqx.nn.Linear(dim, dim, key=key_q)
        self.v = eqx.nn.Linear(dim, dim, key=key_v)
        self.proj = eqx.nn.Linear(dim, dim, key=key_proj)

        self.freqs_cis = precompute_freqs_cis(
            self.head_dim,
            seq_len * 2,
            dtype=self.k.weight.dtype,
        )

    def __call__(self, x: Array):
        T, C = x.shape
        k = jax.vmap(self.k)(x)
        q = jax.vmap(self.q)(x)
        v = jax.vmap(self.v)(x)

        k = k.reshape(T, self.num_heads, C // self.num_heads)
        q = q.reshape(T, self.num_heads, C // self.num_heads)
        v = v.reshape(T, self.num_heads, C // self.num_heads)

        freqs_cis = np.take(self.freqs_cis, T, axis=0)
        q, k = apply_rope(q, k, freqs_cis=freqs_cis, dtype=self.k.weight.dtype)

        k = k.swapaxes(0, 1)
        q = q.swapaxes(0, 1)
        v = v.swapaxes(0, 1)

        mask = np.tril(np.ones((T, T))).reshape((1, T, T))

        attn = q @ k.swapaxes(-2, -1)
        attn = attn * C**-0.5

        attn = np.where(mask == 0, float("-inf"), attn)
        attn = jax.nn.softmax(attn, axis=-1)

        out: Array = attn @ v
        out = out.swapaxes(0, 1).reshape(T, C)

        out = jax.vmap(self.proj)(out)
        return out


class LlamaBlock(eqx.Module):
    attention: Attention
    attn_norm: RMSNorm
    mlp: SwiGLU
    mlp_norm: RMSNorm

    def __init__(self, dim, num_heads, seq_len, *, key):
        super().__init__()
        self.attention = Attention(dim, num_heads, seq_len, key=key)
        self.mlp = eqx.nn.Linear(dim, dim, use_bias=False, key=key)
        self.attn_norm = RMSNorm(dim, key=key)
        self.mlp_norm = RMSNorm(dim, key=key)

    def __call__(self, x: Array):
        x = x + self.attention(self.attn_norm(x))
        x = x + self.mlp(self.mlp_norm(x))
        return x


class LLAMA(eqx.Module):
    dim: int
    num_layers: int
    num_heads: int
    seq_len: int
    vocab_size: int

    embdings: Array
    blocks: tuple[LlamaBlock, ...]
    norm: RMSNorm

    def __init__(self, dim, num_layers, num_heads, seq_len, vocab_size, *, key):
        super().__init__()
        self.emb = jax.random.uniform(
            key,
            shape=(
                vocab_size,
                dim,
            ),
            minval=-1e-4,
            maxval=1e-4,
        )
        self.blocks = tuple(LlamaBlock(dim, num_heads, seq_len, key=key) for _ in range(num_layers))
        self.norm = RMSNorm(dim, key=key)

    def __call__(self, x: Array):
        jax.vmap(lambda x: self.emb[x])(x)
        for block in self.blocks:
            x = block(x)
        x = self.norm(x)
        return self.emb.T @ x


if __name__ == "__main__":
    key = jax.random.PRNGKey(0)
    B, T, C = 3, 5, 256
    # model = Attention(dim=C, num_heads=4, seq_len=T, key=key)
    model = Attention(dim=C, num_heads=4, seq_len=T, key=key)
    x = jax.random.normal(key, (B, T, C))

    out = jax.vmap(model)(x)

    print(out.shape)



import torch


def get_alibi_mask(number_of_heads, max_seq_len):
    """
    https://arxiv.org/abs/2108.12409 \\
    ref: https://www.youtube.com/watch?v=Pp61ShI9VGc \\
    Add this with casulmask then use that mask for efficeny \\
    example:
    ```
    >> mask = casual_mask + get_alibi_mask(number_of_heads,max_seq_len)
    >> wei = softmax(q@k.transpose(-1,-2) + mask[:,:T,:T])
    ```

    """

    nh = number_of_heads
    n = max_seq_len

    rows = torch.arange(1, n + 1).view(1, -1, 1)
    cols = torch.arange(n).view(1, 1, -1)

    matrix = rows - cols
    matrix = torch.where(matrix < 1, (torch.zeros_like(matrix)), matrix)

    matrix = matrix.expand(nh, -1, -1) * -1
    m = 1 / 2 ** (torch.arange(1, nh + 1) / (nh / 8))
    return matrix * m.view(nh, 1, 1)


if __name__ == "__main__":
    print(get_alibi_mask(16, 5))



from __future__ import annotations
import torch
import torch.nn as nn
import torch.nn.functional as F

import math
from dataclasses import dataclass
from ohara.modules.mlp import SwiGLU


@dataclass
class Config:
    vocab_size = 65
    seq_len = 64
    d_model = 128
    num_heads = 4
    num_layers = 4
    dropout = 0.2
    multiple_of = 4
    bias = True


class Attention(nn.Module):
    def __init__(self, model_args: Config):
        super().__init__()
        d_model = model_args.d_model
        self.num_heads = model_args.num_heads
        self.head_dim = model_args.d_model // model_args.num_heads

        self.key = nn.Linear(d_model, d_model)
        self.query = nn.Linear(d_model, d_model)
        self.value = nn.Linear(d_model, d_model)
        self.proj = nn.Linear(d_model, d_model)

        self.attn_dropout = nn.Dropout(model_args.dropout)
        self.res_dropout = nn.Dropout(model_args.dropout)

        self.flash_attn = hasattr(torch.nn.functional, "scaled_dot_product_attention")

    def forward(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        batch, seq_len, d_model = x.shape

        k: torch.Tensor  # type hint for lsp
        q: torch.Tensor  # ignore
        v: torch.Tensor

        k = self.key(x)
        q = self.query(x)
        v = self.value(x)

        k = k.view(
            seq_len, self.num_heads, self.head_dim
        )  # shape = (B, seq_len, num_heads, head_dim)
        q = q.view(seq_len, self.num_heads, self.head_dim)
        v = v.view(seq_len, self.num_heads, self.head_dim)

        k = k.transpose(0, 1)  # shape = (B, num_heads, seq_len, head_dim)
        q = q.transpose(0, 1)
        v = v.transpose(0, 1)

        if self.flash_attn:
            output = torch.nn.functional.scaled_dot_product_attention(
                q,
                k,
                v,  # order impotent
                attn_mask=None,
                dropout_p=self.dropout if self.training else 0.0,
                is_causal=True,
            )
        else:
            attn_mtx = torch.matmul(q, k.transpose(2, 3)) / math.sqrt(self.head_dim)
            attn_mtx = attn_mtx + mask[:, :, :seq_len, :seq_len]
            attn_mtx = F.softmax(attn_mtx.float(), dim=-1).type_as(k)
            attn_mtx = self.attn_dropout(attn_mtx)

            output = torch.matmul(attn_mtx, v)  # (batch, n_head, seq_len, head_dim)

        # restore time as batch dimension and concat heads
        output = output.transpose(1, 2).contiguous().view(batch, seq_len, d_model)

        # final projection into the residual stream
        output = self.proj(output)
        output = self.res_dropout(output)
        return output


class Block(nn.Module):
    def __init__(self, model_args: Config):
        super().__init__()

        self.attn = Attention(model_args)
        self.ff = SwiGLU(
            dim=model_args.d_model,
            multiple_of=model_args.multiple_of,
            dropout=model_args.dropout,
        )

        self.norm1 = nn.LayerNorm(model_args.d_model)
        self.norm2 = nn.LayerNorm(model_args.d_model)

    def forward(self, x, mask):
        x = x + self.attn(self.norm1(x), mask)
        x = x + self.ff(self.norm2(x))
        return x


class GPT(nn.Module):
    def __init__(self, model_args: Config, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.word_emb = nn.Embedding(model_args.vocab_size, model_args.d_model)
        self.pos_emb = nn.Embedding(model_args.seq_len, model_args.d_model)

        self.layers = nn.ModuleList([Block(model_args) for _ in range(model_args.num_layers)])

        self.norm = nn.LayerNorm(model_args.d_model)
        self.vocab_proj = nn.Linear(model_args.d_model, model_args.vocab_size, bias=False)

        if hasattr(torch.nn.functional, "scaled_dot_product_attention"):
            print("WARNING: using slow attention | upgrade pytorch to 2.0 or above")
            mask = torch.full((1, 1, model_args.seq_len, model_args.seq_len), float("-inf"))
            mask = torch.triu(mask, diagonal=1)
            self.register_buffer("mask", mask)
        else:
            self.mask = None

    def forward(self, x):
        x = self.word_emb(x) + self.pos_emb(x)

        for layer in self.layers:
            x = layer(x, self.mask)

        x = self.norm(x)
        x = self.vocab_proj(x)
        return x



from __future__ import annotations

import os
import math
import json
from dataclasses import dataclass


import torch
import torch.nn as nn

from torch import Tensor
from safetensors import safe_open
from ohara.utils.load import download_hf_model

from tqdm import tqdm


@dataclass
class PhiConfig:
    vocab_size: int = 51200
    seq_len: int = 2048
    d_model: int = 2560
    num_heads: int = 32
    num_layers: int = 32
    dropout: float = 0.2
    multiple_of: int = 4
    bias: bool = True
    eps: float = 1e-5
    rotary_dim: float = 0.4


class KVCache:
    def __init__(self, shape, max_seq_length, idx: int | None = None, device=None, dtype=None):
        self.idx = idx
        self.key: torch.Tensor = torch.zeros(shape, device=device, dtype=dtype)
        self.value: torch.Tensor = torch.zeros(shape, device=device, dtype=dtype)
        self.max_seq_length = max_seq_length

    def forward(self, keys: Tensor, values: Tensor, start_pos: Tensor) -> tuple[Tensor, Tensor]:
        bsz, T, _, _ = keys.shape
        # print(f"start_pos={start_pos}, T={T}, bsz={bsz}")
        self.key[:bsz, start_pos : start_pos + T] = keys
        self.value[:bsz, start_pos : start_pos + T] = values
        keys = self.key[:bsz, : start_pos + T]
        values = self.value[:bsz, : start_pos + T]
        return keys, values


class RoPE(nn.Module):
    def __init__(
        self,
        dims: int,
        traditional: bool = False,
        base: float = 10000,
        scale: float = 1.0,
    ):
        super().__init__()

        self.dims = dims
        self.traditional = traditional
        self.base = base
        self.scale = scale

    def _extra_repr(self):
        return f"{self.dims}, traditional={self.traditional}"

    def _compute_rope(self, costheta, sintheta, x) -> Tensor:
        x1 = x[..., : self.dims // 2]
        x2 = x[..., self.dims // 2 : self.dims]
        rx1 = x1 * costheta - x2 * sintheta
        rx2 = x1 * sintheta + x2 * costheta

        if self.dims < x.shape[-1]:
            rx = torch.concatenate([rx1, rx2, x[..., self.dims :]], axis=-1)
        else:
            rx = torch.concatenate([rx1, rx2], axis=-1)

        return rx

    def _compute_traditional_rope(self, costheta, sintheta, x)->Tensor:
        x1 = x[..., ::2]
        x2 = x[..., 1::2]
        rx1 = x1 * costheta - x2 * sintheta
        rx2 = x1 * sintheta + x2 * costheta

        assert self.dims < x.shape[-1], "RoPE doesn't implement partial traditional application"

        rx = torch.cat([rx1[..., None], rx2[..., None]], axis=-1)

        return rx

    def forward(self, x: Tensor, offset: int = 0):
        shape = x.shape
        x = x.reshape(-1, shape[-2], shape[-1])
        N = x.shape[1] + offset
        costheta, sintheta = RoPE.create_cos_sin_theta(
            N,
            self.dims,
            offset=offset,
            base=self.base,
            scale=self.scale,
            dtype=x.dtype,
            device=x.device,
        )

        rope = self._compute_traditional_rope if self.traditional else self._compute_rope
        rx = rope(costheta, sintheta, x)

        return torch.reshape(rx, shape)

    @staticmethod
    def create_cos_sin_theta(
        N: int,
        D: int,
        offset: int = 0,
        base: float = 10000,
        scale: float = 1.0,
        dtype=torch.float32,
        device=None,
    ) -> tuple[Tensor,Tensor]:
        D = D // 2
        positions = torch.arange(offset, N, dtype=dtype, device=device) * scale
        freqs = torch.exp(-torch.arange(0.0, D, dtype=dtype, device=device) * (math.log(base) / D))
        theta = torch.reshape(positions, (-1, 1)) * torch.reshape(freqs, (1, -1))
        return torch.cos(theta), torch.sin(theta)


def new_gelu(x: Tensor) -> Tensor:
    return (
        0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))
    )


class MLP(nn.Module):
    def __init__(self, dim, hidden):
        super().__init__()
        self.fc1 = nn.Linear(dim, hidden)
        self.fc2 = nn.Linear(hidden, dim)

    def forward(self, x: torch.Tensor):
        x = self.fc1(x)
        x = new_gelu(x)
        return self.fc2(x)


class LayerNorm(nn.LayerNorm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: Tensor) -> Tensor:
        return super().forward(x)


class PhiMHA(nn.Module):
    def __init__(self, layer_idx, dim, num_heads, rotary_dim) -> None:
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.layer_idx = layer_idx

        self.k_proj = nn.Linear(dim, dim)
        self.q_proj = nn.Linear(dim, dim)
        self.v_proj = nn.Linear(dim, dim)
        self.dense = nn.Linear(dim, dim)

        self.rope = RoPE(int(rotary_dim * self.head_dim), traditional=False)

    def forward(
        self,
        x: Tensor,
        mask: Tensor = None,
        kv_cache: KVCache | None = None,
        position_ids: Tensor | None = None,
    ) -> Tensor:
        batch_size, seq_length, d_model = x.shape
        # print(f"{batch_size}, {seq_length}, {d_model}")
        k = self.k_proj(x)
        q = self.q_proj(x)
        v = self.v_proj(x)

        k = k.view(batch_size, seq_length, self.num_heads, self.head_dim)
        q = q.view(batch_size, seq_length, self.num_heads, self.head_dim)
        v = v.view(batch_size, seq_length, self.num_heads, self.head_dim)

        if kv_cache is not None:
            k, v = kv_cache.forward(k, v, position_ids)

        k: Tensor = k.transpose(1, 2).to(torch.float32)  # shape = (B, num_heads, seq_len, head_dim)
        q: Tensor = q.transpose(1, 2).to(torch.float32)
        v: Tensor = v.transpose(1, 2).to(torch.float32)

        offset = position_ids if kv_cache else 0

        q = self.rope.forward(q, offset=offset).to(torch.float32)
        k = self.rope.forward(k).to(torch.float32)

        # Finally perform the attention computation
        scale = math.sqrt(1 / q.shape[-1])
        scores = (q @ k.transpose(-1, -2)) * scale

        if mask is not None:
            mask = mask[:, :, :seq_length, :seq_length]
            scores = scores + mask

        scores = torch.softmax(scores, dim=-1).type_as(v)
        output = (scores @ v).type_as(x)
        # output = F.scaled_dot_product_attention(
        #     q,
        #     k,
        #     v,
        #     attn_mask=None,
        #     dropout_p=0,
        #     is_causal=True,
        # ).type_as(x)

        output = output.transpose(1, 2).type_as(x)
        output = output.reshape(batch_size, seq_length, self.dim)

        return self.dense(output)

    def scaled_dot_product_attention(self, q, k, v, mask):
        scale = math.sqrt(self.head_dim)
        scores = (q @ k.transpose(-1, -2)) * scale
        if mask is not None:
            scores = scores + mask
        scores = torch.softmax(scores, dim=-1).type_as(v)
        output = scores @ v
        return output


class Block(nn.Module):
    def __init__(
        self,
        config: PhiConfig,
        block_idx: int | None = None,
    ) -> None:
        super().__init__()

        self.ln = LayerNorm(config.d_model, eps=config.eps)
        self.block_idx = block_idx

        self.mixer = PhiMHA(block_idx, config.d_model, config.num_heads, config.rotary_dim)
        self.mlp = MLP(config.d_model, config.multiple_of * config.d_model)

    def forward(
        self,
        x: torch.FloatTensor,
        mask: torch.BoolTensor | None = None,
        kv_cache: list[KVCache] | None = None,
        position_ids: torch.LongTensor | None = None,
    ) -> torch.FloatTensor:
        residual = x
        x = self.ln(x)
        return self.mixer(x, mask, kv_cache, position_ids) + self.mlp(x) + residual


class Phi(nn.Module):
    def __init__(self, config: PhiConfig) -> None:
        super().__init__()
        self.config = config
        self.wte = nn.Embedding(config.vocab_size, config.d_model)
        self.layers = nn.ModuleList([Block(config, i) for i in range(config.num_layers)])

        self.ln = LayerNorm(config.d_model, eps=config.eps)
        self.lm_head = nn.Linear(config.d_model, config.vocab_size)
        self.loss_fn = nn.CrossEntropyLoss()

        mask = torch.full((1, 1, config.seq_len, config.seq_len), float("-inf"))

        (
            1,
            config.seq_len,
            config.num_heads,
            config.d_model // config.num_heads,
        )

        mask = torch.triu(mask, diagonal=1)
        self.register_buffer("mask", mask)

    def forward(self, x, kv_cache: list[KVCache] | None = None, position_ids=None):
        mask = self.mask
        if kv_cache is not None:
            x = x[:, position_ids:]
            mask = None
        # print(f"{x.shape=} {kv_cache=}")
        x = self.wte(x)
        # print(f"{x.sum(dim=-1)=}")
        cache = None
        for idx, layer in enumerate(self.layers):
            if kv_cache is not None:
                cache = kv_cache[idx]
            x = layer(x.to(self.wte.weight.dtype), mask, cache, position_ids=position_ids)

        x = self.ln(x)
        x = self.lm_head(x)
        return x

    def loss(self, logits: Tensor, labels: Tensor):
        loss = self.loss_fn(logits.view(-1, logits.size(-1)), labels.view(-1))
        return loss

    def build_kv_cache(self) -> list[KVCache]:
        shape = (
            1,
            self.config.seq_len,
            self.config.num_heads,
            self.config.d_model // self.config.num_heads,
        )
        kv_cache = []
        dtype = self.wte.weight.dtype
        device = self.wte.weight.device

        for idx in range(self.config.num_layers):
            kv_cache.append(KVCache(shape, self.config.seq_len, idx, device=device, dtype=dtype))
        return kv_cache

    @staticmethod
    def from_pretrained(name: str) -> nn.Module:
        config = PhiConfig()
        with torch.device("cuda"):
            model = Phi(config).half()
        # return model

        path_name = download_hf_model(name)
        with open(os.path.join(path_name, "config.json"), encoding="utf-8") as f:
            json.load(f)

        files = ["model-00001-of-00002.safetensors", "model-00002-of-00002.safetensors"]
        weights = {}
        for f in files:
            w = safe_open(os.path.join(path_name, f), framework="pt", device="cpu")
            for k in w.keys():
                # print(k)
                weights[k] = w.get_tensor(k)

        assert model.wte.weight.data.shape == weights["model.embed_tokens.weight"].shape
        model.wte.weight.data = weights["model.embed_tokens.weight"]

        assert model.lm_head.weight.data.shape == weights["lm_head.weight"].shape
        model.lm_head.weight.data = weights["lm_head.weight"]

        assert model.lm_head.bias.data.shape == weights["lm_head.bias"].shape
        model.lm_head.bias.data = weights["lm_head.bias"]

        assert model.ln.weight.data.shape == weights["model.final_layernorm.weight"].shape
        model.ln.weight.data = weights["model.final_layernorm.weight"]

        assert model.ln.bias.data.shape == weights["model.final_layernorm.bias"].shape
        model.ln.bias.data = weights["model.final_layernorm.bias"]

        for idx, layer in tqdm(enumerate(model.layers), total=len(model.layers)):
            layer: Block

            assert (
                layer.mlp.fc1.weight.data.shape
                == weights[f"model.layers.{idx}.mlp.fc1.weight"].shape
            )
            layer.mlp.fc1.weight.data = weights[f"model.layers.{idx}.mlp.fc1.weight"]

            assert (
                layer.mlp.fc1.bias.data.shape == weights[f"model.layers.{idx}.mlp.fc1.bias"].shape
            )
            layer.mlp.fc1.bias.data = weights[f"model.layers.{idx}.mlp.fc1.bias"]

            assert (
                layer.mlp.fc2.weight.data.shape
                == weights[f"model.layers.{idx}.mlp.fc2.weight"].shape
            )
            layer.mlp.fc2.weight.data = weights[f"model.layers.{idx}.mlp.fc2.weight"]

            assert (
                layer.mlp.fc2.bias.data.shape == weights[f"model.layers.{idx}.mlp.fc2.bias"].shape
            )
            layer.mlp.fc2.bias.data = weights[f"model.layers.{idx}.mlp.fc2.bias"]

            assert (
                layer.mixer.q_proj.weight.data.shape
                == weights[f"model.layers.{idx}.self_attn.q_proj.weight"].shape
            )
            layer.mixer.q_proj.weight.data = weights[f"model.layers.{idx}.self_attn.q_proj.weight"]

            assert (
                layer.mixer.q_proj.bias.data.shape
                == weights[f"model.layers.{idx}.self_attn.q_proj.bias"].shape
            )
            layer.mixer.q_proj.bias.data = weights[f"model.layers.{idx}.self_attn.q_proj.bias"]

            assert (
                layer.mixer.k_proj.weight.data.shape
                == weights[f"model.layers.{idx}.self_attn.k_proj.weight"].shape
            )
            layer.mixer.k_proj.weight.data = weights[f"model.layers.{idx}.self_attn.k_proj.weight"]

            assert (
                layer.mixer.k_proj.bias.data.shape
                == weights[f"model.layers.{idx}.self_attn.k_proj.bias"].shape
            )
            layer.mixer.k_proj.bias.data = weights[f"model.layers.{idx}.self_attn.k_proj.bias"]

            assert (
                layer.mixer.v_proj.weight.data.shape
                == weights[f"model.layers.{idx}.self_attn.v_proj.weight"].shape
            )
            layer.mixer.v_proj.weight.data = weights[f"model.layers.{idx}.self_attn.v_proj.weight"]

            assert (
                layer.mixer.v_proj.bias.data.shape
                == weights[f"model.layers.{idx}.self_attn.v_proj.bias"].shape
            )
            layer.mixer.v_proj.bias.data = weights[f"model.layers.{idx}.self_attn.v_proj.bias"]

            assert (
                layer.mixer.dense.weight.data.shape
                == weights[f"model.layers.{idx}.self_attn.dense.weight"].shape
            )
            layer.mixer.dense.weight.data = weights[f"model.layers.{idx}.self_attn.dense.weight"]

            assert (
                layer.mixer.dense.bias.data.shape
                == weights[f"model.layers.{idx}.self_attn.dense.bias"].shape
            )
            layer.mixer.dense.bias.data = weights[f"model.layers.{idx}.self_attn.dense.bias"]

            assert (
                layer.ln.weight.data.shape
                == weights[f"model.layers.{idx}.input_layernorm.weight"].shape
            )
            layer.ln.weight.data = weights[f"model.layers.{idx}.input_layernorm.weight"]

            assert (
                layer.ln.bias.data.shape
                == weights[f"model.layers.{idx}.input_layernorm.bias"].shape
            )
            layer.ln.bias.data = weights[f"model.layers.{idx}.input_layernorm.bias"]

        return model



import torch
from torch import Tensor


def repeat_for_grouped_query(x: Tensor, num_queries: int):
    # assuming shape batch, seq_len, num_heads, head_dim
    x = torch.repeat_interleave(x, num_queries, dim=2)



import math

from dataclasses import dataclass
from typing import Any


@dataclass
class Scheduler:
    learning_rate: float = 3e-4

    def __call__(self, *args: Any, **kwds: Any) -> float:
        return self.learning_rate


@dataclass
class CosineScheduler:
    learning_rate: float = 3e-4  # Karpathy constant
    min_lr: float = 3e-5  # 1/10 of lr as per chichilla papaer
    warmup_iters: int = 1000
    max_iters: int = 100_0000

    def __call__(self, iteration) -> float:
        if iteration < self.warmup_iters:
            return self.learning_rate * iteration / self.warmup_iters

        if iteration > self.max_iters:
            return self.min_lr

        decay_ratio = (iteration - self.warmup_iters) / (self.max_iters - self.warmup_iters)
        assert 0 <= decay_ratio <= 1

        coeff = 0.5 * (1.0 + math.cos(math.pi * decay_ratio))
        return self.min_lr + coeff * (self.learning_rate - self.min_lr)


# TODO: jagged cosine learning rate for relora
# I belive parameter efficiency pretraning is possibe only one way to find out
# relora would be good start

if __name__ == "__main__":
    scheduler = CosineScheduler(learning_rate=0.1, min_lr=0.001, warmup_iters=5, max_iters=100)

    print(scheduler)



from __future__ import annotations

import os
import math
import json
from dataclasses import dataclass


import torch
import torch.nn as nn
import torch.nn.functional as F


from torch import Tensor
from safetensors import safe_open
from ohara.utils.load import download_hf_model

from tqdm import tqdm
from typing import Tuple

from ohara.modules.mlp import GEGLU
from ohara.modules.norm import RMSNorm
from ohara.embedings_pos.rotatry import precompute_freqs_cis
from ohara.embedings_pos.rotatry import apply_rope


@dataclass
class GemmaConfig:
    vocab_size: int = 51200
    seq_len: int = 2048
    d_model: int = 2048
    intermediate_size = 16 * 2048
    num_heads: int = 32
    num_kv_heads: int = 1
    num_layers: int = 32
    dropout: float = 0.2
    multiple_of: int = 4
    bias: bool = True
    eps: float = 1e-5
    rotary_dim: float = 0.4


class GemmaAttention(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        num_kv_heads: int,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.num_kv_heads = num_kv_heads

        assert self.num_heads % self.num_kv_heads == 0
        self.num_queries_per_kv = self.num_heads // self.num_kv_heads

        self.d_model = d_model
        self.head_dim = self.d_model // num_heads

        self.q_size = self.num_heads * self.head_dim
        self.kv_size = self.num_kv_heads * self.head_dim

        self.scaling: float = self.head_dim**-0.5

        self.qkv_proj = nn.Linear(
            self.d_model, (self.num_heads + 2 * self.num_kv_heads) * self.head_dim
        )

        self.o_proj = nn.Linear(self.num_heads * self.head_dim, self.d_model)

    def forward(
        self,
        x: torch.Tensor,
        freqs_cis: torch.Tensor,
        mask: torch.Tensor = None,
    ) -> torch.Tensor:
        input_shape = x.shape
        assert len(input_shape) == 3

        batch_size, seq_len, d_model = x.shape

        qkv: Tensor = self.qkv_proj(x)
        q, k, v = qkv.split([self.q_size, self.kv_size, self.kv_size], dim=-1)

        q: Tensor = q.view(batch_size, -1, self.num_heads, self.head_dim)
        k: Tensor = k.view(batch_size, -1, self.num_kv_heads, self.head_dim)
        v: Tensor = v.view(batch_size, -1, self.num_kv_heads, self.head_dim)

        # Positional embedding.
        q = apply_rope(q, freqs_cis=freqs_cis)
        k = apply_rope(k, freqs_cis=freqs_cis)

        # TODO: add code for kv chache

        # Grouped Query Attention
        if self.num_kv_heads != self.num_heads:
            key = torch.repeat_interleave(k, self.num_queries_per_kv, dim=2)
            value = torch.repeat_interleave(v, self.num_queries_per_kv, dim=2)

        q = q.transpose(1, 2)
        k = key.transpose(1, 2)
        v = value.transpose(1, 2)

        attn_mtx = torch.matmul(q, k.transpose(2, 3)) * self.scaling
        attn_mtx = attn_mtx + mask
        attn_mtx = F.softmax(attn_mtx.float(), dim=-1).type_as(q)

        output = torch.matmul(attn_mtx, v)

        output = output.transpose(1, 2).contiguous().view(batch_size, seq_len, -1)
        output = self.o_proj(output)
        return output


class Block(nn.Module):
    def __init__(
        self,
        config: GemmaConfig,
    ):
        super().__init__()
        self.self_attn = GemmaAttention(
            d_model=config.d_model,
            num_heads=config.num_heads,
            num_kv_heads=config.num_kv_heads,
        )
        self.mlp = GEGLU(
            dim=config.d_model,
            hidden_dim=config.intermediate_size,
        )
        self.ln1 = RMSNorm(config.d_model, eps=config.eps)
        self.ln2 = RMSNorm(config.d_model, eps=config.eps)

    def forward(
        self,
        x: torch.Tensor,
        freqs_cis: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        x = x + self.self_attn(x=self.ln1(x), freqs_cis=freqs_cis, mask=mask)
        x = x + self.mlp(self.ln2(x))

        return x


class Gemma(nn.Module):
    def __init__(self, model_args: GemmaConfig, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.config = model_args

        self.token_emb = nn.Embedding(model_args.vocab_size, model_args.d_model)

        self.layers = nn.ModuleList([Block(model_args) for _ in range(model_args.num_layers)])

        self.norm = nn.LayerNorm(model_args.d_model)
        self.vocab_proj = nn.Linear(model_args.d_model, model_args.vocab_size, bias=False)

        self.token_emb.weight = self.vocab_proj.weight

        self.cos, self.sin = precompute_freqs_cis(
            model_args.d_model // model_args.num_heads, model_args.seq_len * 2
        )

        if hasattr(torch.nn.functional, "scaled_dot_product_attention"):
            print("WARNING: using slow attention | upgrade pytorch to 2.0 or above")
            mask = torch.full((1, 1, model_args.seq_len, model_args.seq_len), float("-inf"))
            mask = torch.triu(mask, diagonal=1)
            self.register_buffer("mask", mask)
        else:
            self.mask = None

    def forward(self, x: torch.Tensor):
        batch, seqlen = x.shape
        x = self.token_emb(x)

        device = self.token_emb.weight.device
        freqs_cis = self.cos[:seqlen].to(device), self.sin[:seqlen].to(device)

        for layer in self.layers:
            x = layer(x, self.mask, freqs_cis)

        x = self.norm(x)
        x = self.vocab_proj(x)
        return x



import time
import torch
import torch.nn as nn
from tqdm import tqdm

from torch import Tensor
from typing import Optional

from transformers import AutoTokenizer


class Inference:
    def __init__(
        self,
        model: nn.Module,
        tokenizer: AutoTokenizer,
        device: str = None,
        temperature: float = 1.0,
        top_p: float = 0.0,
        max_tokens: int = 500,
        use_kv_cache: bool = True,
    ):
        self.model = model.to(device).eval()
        self.tokenizer = tokenizer
        self.device = device
        self.default_temperature = temperature
        self.default_top_p = top_p
        self.max_tokens = max_tokens
        self.use_kv_cache = use_kv_cache
        if use_kv_cache and hasattr(self.model, "build_kv_cache"):
            self.kv_cache = self.model.build_kv_cache()
        else:
            self.kv_cache = None
        if device is None:
            self.device = torch.device("cpu")
            if torch.cuda.is_available():
                self.device = torch.device("cuda")

    @staticmethod
    @torch.inference_mode()
    def sampler(logits, temperature=1, top_p=0.0) -> torch.Tensor:
        logits = logits[:, -1]
        if temperature == 1:
            return torch.argmax(logits, dim=-1, keepdim=True)
        logits = logits / temperature

        probs = torch.softmax(logits, dim=-1)

        probs_sort, probs_idx = torch.sort(probs, dim=-1, descending=True)  # (B, vocab_size)

        probs_sum = torch.cumsum(probs_sort, dim=-1)

        mask = probs_sum - probs_sort > top_p

        probs_sort[mask] = 0.0

        probs_sort.div_(probs_sort.sum(dim=-1, keepdim=True))

        next_token = torch.multinomial(probs_sort, num_samples=1)

        next_token = torch.gather(probs_idx, -1, next_token)
        return next_token

    def generate(
        self,
        prompt: str,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream: bool = True,
    ):
        if temperature is None:
            temperature = self.default_temperature
        if top_p is None:
            top_p = self.default_top_p
        if hasattr(self.model, "build_kv_cache"):
            self.kv_cache = self.model.build_kv_cache()
        else:
            self.kv_cache = None
        device = torch.device("cuda")
        inputs = self.tokenizer.encode(prompt)
        inputs = torch.tensor(inputs).reshape(1, -1).to(device)
        input_pos = 0
        start_time = time.time()
        with torch.no_grad():
            if stream:
                print(self.tokenizer.decode(inputs.tolist()[0]), end="")
            for _ in range(self.max_tokens):
                logits = (
                    self.model(inputs, self.kv_cache, input_pos)
                    if self.use_kv_cache
                    else self.model(inputs)
                )
                next_token = self.sampler(logits, temperature=temperature, top_p=top_p)
                if next_token[:, -1:].item() == self.tokenizer.eos_token_id:
                    break
                inputs = torch.cat((inputs, next_token[:, -1:]), dim=-1)
                input_pos = inputs.shape[1] - 1
                if stream:
                    print(self.tokenizer.decode(inputs.tolist()[0][-1]), end="", flush=True)
            end_time = time.time()
        if stream:
            print(f"\nTime: {end_time - start_time}s")
        return self.tokenizer.decode(inputs.tolist()[0][-1])


if __name__ == "__main__":
    torch.manual_seed(0)

    allow_cuda = False
    device = "cuda" if torch.cuda.is_available() and allow_cuda else "cpu"

    prompts = [
        "Simply put, the theory of relativity states that ",
        "If Google was an Italian company founded in Milan, it would",
        # Few shot promt
        """Translate English to French:

        sea otter => loutre de mer
        peppermint => menthe poivrée
        plush girafe => girafe peluche
        cheese =>""",
        # Zero shot prompt
        """Tell me if the following person is actually Doraemon disguised as human:
        Name: Umar Jamil
        Decision:
        """,
    ]

    model = LLaMA.build(
        checkpoints_dir="llama-2-7b/",
        tokenizer_path="tokenizer.model",
        load_model=True,
        max_seq_len=1024,
        max_batch_size=len(prompts),
        device=device,
    )

    out_tokens, out_texts = model.text_completion(prompts, max_gen_len=64)
    assert len(out_texts) == len(prompts)
    for i in range(len(out_texts)):
        print(f"{out_texts[i]}")
        print("-" * 50)



from __future__ import annotations

import torch
import torch.nn as nn


class XPos(nn.Module):
    def __init__(self, dim, num_heads):
        """
        https://arxiv.org/abs/2212.10554
        """
        super().__init__()

        angle = 1.0 / (10000 ** torch.linspace(0, 1, dim // num_heads // 2))
        angle = angle.unsqueeze(-1).repeat(1, 2).flatten()
        decay = torch.log(1 - 2 ** (-5 - torch.arange(num_heads, dtype=torch.float)))
        self.register_buffer("angle", angle)
        self.register_buffer("decay", decay)

    def forward(self, slen, recurrent=False):
        # todo chukwire retation
        if recurrent:
            sin = torch.sin(self.angle * (slen - 1))
            cos = torch.cos(self.angle * (slen - 1))
            retention_rel_pos = ((cos, sin), self.decay.exp())
        else:
            index = torch.arange(slen).to(self.decay)
            sin = torch.sin(index[:, None] * self.angle[None, :])
            cos = torch.cos(index[:, None] * self.angle[None, :])
            mask = torch.tril(torch.ones(slen, slen).to(self.decay))
            mask = torch.masked_fill(index[:, None] - index[None, :], ~mask.bool(), float("inf"))
            mask = torch.exp(mask * self.decay[:, None, None])
            mask = torch.nan_to_num(mask)
            mask = mask / mask.sum(dim=-1, keepdim=True).sqrt()
            retention_rel_pos = ((cos, sin), mask)

        return retention_rel_pos


if __name__ == "__main__":
    xpos = XPos(64, 4)
    ((cos, sin), decay) = xpos.forward(8)
    print(decay)

    for i in range(1, 9):
        ((sin, cos), decay) = xpos.forward(8, recurrent=True)
        print(i, decay)



import torch
import torch.nn.functional as F
from torch import Tensor
import time

import jax.numpy as jnp
from jax.nn import softmax
import jax
from jax import lax

torch.manual_seed(2)


def sliding_window_attention(
    q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, window_size: int = 16
):
    B, T, C = q.shape
    out = torch.zeros(B, T, C, device=q.device)
    for i in range(T):
        start = max(0, i - window_size + 1)
        end = min(T, i + 1)
        key = k[:, start:end, :]
        query = q[:, i, :].unsqueeze(1)
        value = v[:, start:end, :]

        # Compute scaled dot-product attention
        wei = query @ key.transpose(-1, -2)
        wei = F.softmax(wei, dim=-1)  # Apply softmax to weights
        out[:, i, :] = (wei @ value).squeeze(1)  # Multiply weights by values

    return out


def sliding_window_attention_with_mask(q: Tensor, k: Tensor, v: Tensor, window_size: int = 16):
    B, T, C = q.shape
    out = torch.zeros(T, C)
    mask = torch.zeros(T, T)
    for i in range(T):
        for j in range(T):
            if j > i:
                mask[i, j] = float("-inf")
            if i - j > window_size - 1:
                mask[i, j] = float("-inf")

    wei = q @ k.transpose(-1, -2)
    wei = wei + mask
    wei = F.softmax(wei, dim=-1)

    out = wei @ v
    return out


B, T, C = 5, 1000, 8

q = torch.rand(B, T, C)
k = torch.rand(B, T, C)
v = torch.rand(B, T, C)

window_size = 3

# out = sliding_window_attention(q,k,v,window_size=window_size)
# print(out.shape)
sliding_window_attention = torch.compile(sliding_window_attention)
out = sliding_window_attention(q, k, v, window_size=window_size)
ITERS = 1000
start: float = time.perf_counter()
for _ in range(ITERS):
    out = sliding_window_attention(q, k, v, window_size=window_size)
end: float = time.perf_counter()
print(f"time:{(end-start)/ITERS}")


start: float = time.perf_counter()
for _ in range(ITERS):
    out = sliding_window_attention_with_mask(q, k, v, window_size=window_size)
end: float = time.perf_counter()
print(f"time:{(end-start)/ITERS}")
print(out)

exit(0)


def sliding_window_attention_jax(q, k, v, window_size=16):
    T, C = q.shape
    out = jnp.zeros((T, C))
    max_slice_size = window_size  # Set the maximum slice size

    def attention_step(i, out):
        start = lax.max(0, i - window_size + 1)
        end = lax.min(T, i + 1)
        slice_size = end - start

        # Use a fixed-size slice with masking
        k_slice = lax.dynamic_slice(k, (start, 0), (max_slice_size, C))
        v_slice = lax.dynamic_slice(v, (start, 0), (max_slice_size, C))

        # Create a mask for valid positions
        valid_mask = jnp.arange(max_slice_size) < slice_size

        # Apply the mask
        k_slice_masked = k_slice * valid_mask[:, None]
        v_slice_masked = v_slice * valid_mask[:, None]

        weights = softmax(jnp.dot(k_slice_masked, q[i]), axis=-1).reshape(-1, 1)
        weighted_sum = jnp.sum(weights * v_slice_masked, axis=0)
        return out.at[i].set(weighted_sum)

    out = lax.fori_loop(0, T, attention_step, out)
    return out


# Dummy input tensors for testing the function
B, T, C = 10, 50, 128  # Example dimensions
q = jnp.ones((B, T, C))  # Query tensor
k = jnp.ones((B, T, C))  # Key tensor
v = jnp.ones((B, T, C))  # Value tensor

# Call the JAX function with the dummy inputs
output_jax = jax.vmap(sliding_window_attention_jax)(q, k, v)
print(output_jax.shape)  # Display the shape of the output for verification



import sys
import time
import math
from datetime import datetime

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from torch import Tensor
from typing import Any, Callable

from ohara.models.llama import LLAMA, Config
from ohara.lr_scheduler import CosineScheduler
from ohara.dataset import PreTokenizedDataset
from ohara.utils import auto_accelerator, model_summary, BetterCycle

from torch.utils.data import DataLoader
from transformers import AutoTokenizer

import lightning as L
from lightning.pytorch.loggers import WandbLogger
from lightning.fabric.loggers import TensorBoardLogger

import wandb
from rich import print, traceback

traceback.install()



class Trainer:
    def __init__(
        self,
        fabric: L.Fabric,
        model: nn.Module,
        optimizer: optim.Optimizer,
        train_dataloader: DataLoader,
        val_dataloader: DataLoader,
        get_lr: Callable[[int], float],
        micro_batch: int,
        max_iters: int,
        eval_iters: int,
        save_ckpt_iters: int,
        ignore_index: int = -1,
        push_to_hub: bool = False,
        model_name: str = "",
    ):
        self.fabric = fabric
        self.model = model
        self.optimizer = optimizer
        self.train_dataloader = BetterCycle(iter(train_dataloader))
        self.val_dataloader = BetterCycle(iter(val_dataloader))
        self.get_lr = get_lr
        self.micro_batch = micro_batch
        self.max_iters = max_iters
        self.eval_iters = eval_iters
        self.save_ckpt_iters = save_ckpt_iters
        self.ignore_index = ignore_index
        self.push_to_hub = push_to_hub
        self.model_name = model_name

        (data, target) = next(self.val_dataloader)
        self.tokens_per_iter = int(math.prod(data.shape) * micro_batch)

    @torch.no_grad()
    def calculate_loss(self, dataloader: DataLoader, num_batches: int) -> torch.Tensor:
        self.model.eval()
        losses = torch.zeros(num_batches, device=self.fabric.device)
        for idx, (data, target) in enumerate(dataloader):
            if idx >= num_batches:
                break
            logits: torch.Tensor = self.model(data)
            loss = F.cross_entropy(
                logits.view(-1, logits.size(-1)),
                target.reshape(-1),
                ignore_index=self.ignore_index,
            )
            losses[idx] = loss.item()
        return losses.mean()

    def log_function(self, idx: int, lr: float, elapsed_time: float) -> None:
        train_loss = self.calculate_loss(self.train_dataloader, 100)
        val_loss = self.calculate_loss(self.val_dataloader, 100)
        
        print(f"iter: {idx} | train_loss: {train_loss:.4f} | val_loss: {val_loss:.4f} | lr: {lr:e} | time: {elapsed_time:.4f}s")
        
        try:
            self.fabric.log_dict(
                {
                    "training_loss": train_loss,
                    "validation_loss": val_loss,
                    "iter": idx,
                    "tokens": idx * self.tokens_per_iter,
                    "lr": lr,
                    "time": elapsed_time,
                },
                step=idx,
            )
        except Exception as e:
            print(f"Error logging: {e}")

    def train(self, start_iter: int = 0):
        self.fabric.launch()
        # sanity test
        self.calculate_loss(self.val_dataloader, 5)

        idx: int = start_iter
        while True:
            if idx >= self.max_iters:
                break
            idx += 1
            start_time: float = time.perf_counter()

            lr = self.get_lr(idx)
            for param_group in self.optimizer.param_groups:
                param_group["lr"] = lr

            micro_batch_loss = 0
            for _ in range(self.micro_batch):
                (data, target) = next(self.train_dataloader)
                with self.fabric.no_backward_sync(self.model, enabled=_ < self.micro_batch - 1):
                    logits: torch.Tensor = self.model(data)
                    loss = F.cross_entropy(
                        logits.view(-1, logits.size(-1)),
                        target.view(-1),
                        ignore_index=self.ignore_index,
                    )
                    loss = loss / self.micro_batch
                    micro_batch_loss += loss.item()
                    self.fabric.backward(loss)

            self.optimizer.step()
            self.optimizer.zero_grad()

            curr_time: float = time.perf_counter()
            elapsed_time: float = curr_time - start_time
            print(
                f"iter: {idx} | loss: {micro_batch_loss:.4f} | lr: {lr:e} | time: {elapsed_time:.4f}s"
            )

            if idx % self.eval_iters == 0:
                self.model.eval()
                self.log_function(idx, lr, elapsed_time)
                self.model.train()

            if idx % self.save_ckpt_iters == 0:
                state = {"model": self.model.state_dict(), "optimizer": self.optimizer.state_dict(), "idx": idx, "lr": lr}
                self.fabric.save("./ckpt/model.pt", state)
                self.model.config.ckpt_iter = idx
                if self.push_to_hub:
                    self.model.push_to_hub(self.model_name, commit_message=f"checkpoint iter: {idx}")


def main():
    # wandb
    project_name: str = "Ohara-LLAMA-Trainer"

    # dataset and tokenizer
    pretrained_model: str = "microsoft/phi-2"
    dataset_name: str = "roneneldan/TinyStories"

    # learning_rate
    learning_rate: float = 5e-4
    min_lr: float = 0.0

    warmup_iters: int = 1000
    max_iters: int = 100_000
    eval_iters: int = 100

    # bactch size
    batch_size: int = 32
    micro_batch: int = 4

    # Model Args
    d_model: int = 128
    seq_len: int = 256
    num_layers: int = 4
    num_heads: int = 4
    multiple_of: int = 4
    max_length: int = 256

    # system
    compile_model: bool = not (sys.platform != "darwin")
    device: torch.device = auto_accelerator()  # select accelerator eg cuda, mps
    dtype: torch.dtype = torch.float32  # We will use fancy dtypes in future

    logger: Any = wandb.init(project=project_name)

    tokenizer: AutoTokenizer = AutoTokenizer.from_pretrained(pretrained_model)

    config: Config = Config(
        vocab_size=tokenizer.vocab_size,
        d_model=d_model,
        seq_len=seq_len,
        num_layers=num_layers,
        num_heads=num_heads,
        multiple_of=multiple_of,
    )

    model: nn.Module = LLAMA(config).to(device)
    if compile_model:
        model = torch.compile(model)

    print("-" * 100)
    print(model)
    print(model_summary(model))
    print("-" * 100)

    train_ds: PreTokenizedDataset = PreTokenizedDataset(
        dataset_name=dataset_name,
        tokenizer=tokenizer,
        split="train",
        max_length=max_length,
    )
    test_ds: PreTokenizedDataset = PreTokenizedDataset(
        dataset_name=dataset_name,
        tokenizer=tokenizer,
        split="validation",
        max_length=max_length,
    )

    train_dataloader: DataLoader = DataLoader(train_ds, batch_size=batch_size)
    val_dataloader: DataLoader = DataLoader(test_ds, batch_size=batch_size)

    optimizer: optim.AdamW = optim.AdamW(model.parameters(), lr=learning_rate)
    scheduler: CosineScheduler = CosineScheduler(
        learning_rate=learning_rate,
        min_lr=min_lr,
        warmup_iters=warmup_iters,
        max_iters=max_iters,
    )

    trainer: Trainer = Trainer(
        fabric=L.Fabric(accelerator="auto", devices="auto", precision="bf16-mixed"),
        model=model,
        optimizer=optimizer,
        train_dataloader=train_dataloader,
        val_dataloader=val_dataloader,
        get_lr=scheduler,
        micro_batch=micro_batch,
        max_iters=max_iters,
        eval_iters=eval_iters,
        save_ckpt_iters=1000,
        ignore_index=tokenizer.pad_token_id,
        push_to_hub=False,
        model_name="",
    )

    trainer.train()

    wandb.finish()


if __name__ == "__main__":
    main()



import random
from datetime import datetime


pokemons_names = [
    "bulbasaur",
    "ivysaur",
    "venusaur",
    "charmander",
    "charmeleon",
    "charizard",
    "squirtle",
    "wartortle",
    "blastoise",
    "caterpie",
    "metapod",
    "butterfree",
    "weedle",
    "kakuna",
    "beedrill",
    "pidgey",
    "pidgeotto",
    "pidgeot",
    "rattata",
    "raticate",
    "spearow",
    "fearow",
    "ekans",
    "arbok",
    "pikachu",
    "raichu",
    "sandshrew",
    "sandslash",
    "nidorina",
    "nidoqueen",
    "nidorino",
    "nidoking",
    "clefairy",
    "clefable",
    "vulpix",
    "ninetales",
    "jigglypuff",
    "wigglytuff",
    "zubat",
    "golbat",
    "oddish",
    "gloom",
    "vileplume",
    "paras",
    "parasect",
    "venonat",
    "venomoth",
    "diglett",
    "dugtrio",
    "meowth",
    "persian",
    "psyduck",
    "golduck",
    "mankey",
    "primeape",
    "growlithe",
    "arcanine",
    "poliwag",
    "poliwhirl",
    "poliwrath",
    "abra",
    "kadabra",
    "alakazam",
    "machop",
    "machoke",
    "machamp",
    "bellsprout",
    "weepinbell",
    "victreebel",
    "tentacool",
    "tentacruel",
    "geodude",
    "graveler",
    "golem",
    "ponyta",
    "rapidash",
    "slowpoke",
    "slowbro",
    "magnemite",
    "magneton",
    "doduo",
    "dodrio",
    "seel",
    "dewgong",
    "grimer",
    "muk",
    "shellder",
    "cloyster",
    "gastly",
    "haunter",
    "gengar",
    "onix",
    "drowzee",
    "hypno",
    "krabby",
    "kingler",
    "voltorb",
    "electrode",
    "exeggcute",
    "exeggutor",
    "cubone",
    "marowak",
    "hitmonlee",
    "hitmonchan",
    "lickitung",
    "koffing",
    "weezing",
    "rhyhorn",
    "rhydon",
    "chansey",
    "tangela",
    "kangaskhan",
    "horsea",
    "seadra",
    "goldeen",
    "seaking",
    "staryu",
    "starmie",
    "scyther",
    "jynx",
    "electabuzz",
    "magmar",
    "pinsir",
    "tauros",
    "magikarp",
    "gyarados",
    "lapras",
    "ditto",
    "eevee",
    "vaporeon",
    "jolteon",
    "flareon",
    "porygon",
    "omanyte",
    "omastar",
    "kabuto",
    "kabutops",
    "aerodactyl",
    "snorlax",
    "articuno",
    "zapdos",
    "moltres",
    "dratini",
    "dragonair",
    "dragonite",
    "mewtwo",
    "mew",
    "chikorita",
    "bayleef",
    "meganium",
    "cyndaquil",
    "quilava",
    "typhlosion",
    "totodile",
    "croconaw",
    "feraligatr",
    "sentret",
    "furret",
    "hoothoot",
    "noctowl",
    "ledyba",
    "ledian",
    "spinarak",
    "ariados",
    "crobat",
    "chinchou",
    "lanturn",
    "pichu",
    "cleffa",
    "igglybuff",
    "togepi",
    "togetic",
    "natu",
    "xatu",
    "mareep",
    "flaaffy",
    "ampharos",
    "bellossom",
    "marill",
    "azumarill",
    "sudowoodo",
    "politoed",
    "hoppip",
    "skiploom",
    "jumpluff",
    "aipom",
    "sunkern",
    "sunflora",
    "yanma",
    "wooper",
    "quagsire",
    "espeon",
    "umbreon",
    "murkrow",
    "slowking",
    "misdreavus",
    "unown",
    "wobbuffet",
    "girafarig",
    "pineco",
    "forretress",
    "dunsparce",
    "gligar",
    "steelix",
    "snubbull",
    "granbull",
    "qwilfish",
    "scizor",
    "shuckle",
    "heracross",
    "sneasel",
    "teddiursa",
    "ursaring",
    "slugma",
    "magcargo",
    "swinub",
    "piloswine",
    "corsola",
    "remoraid",
    "octillery",
    "delibird",
    "mantine",
    "skarmory",
    "houndour",
    "houndoom",
    "kingdra",
    "phanpy",
    "donphan",
    "stantler",
    "smeargle",
    "tyrogue",
    "hitmontop",
    "smoochum",
    "elekid",
    "magby",
    "miltank",
    "blissey",
    "raikou",
    "entei",
    "suicune",
    "larvitar",
    "pupitar",
    "tyranitar",
    "lugia",
    "celebi",
    "treecko",
    "grovyle",
    "sceptile",
    "torchic",
    "combusken",
    "blaziken",
    "mudkip",
    "marshtomp",
    "swampert",
    "poochyena",
    "mightyena",
    "zigzagoon",
    "linoone",
    "wurmple",
    "silcoon",
    "beautifly",
    "cascoon",
    "dustox",
    "lotad",
    "lombre",
    "ludicolo",
    "seedot",
    "nuzleaf",
    "shiftry",
    "taillow",
    "swellow",
    "wingull",
    "pelipper",
    "ralts",
    "kirlia",
    "gardevoir",
    "surskit",
    "masquerain",
    "shroomish",
    "breloom",
    "slakoth",
    "vigoroth",
    "slaking",
    "nincada",
    "ninjask",
    "shedinja",
    "whismur",
    "loudred",
    "exploud",
    "makuhita",
    "hariyama",
    "azurill",
    "nosepass",
    "skitty",
    "delcatty",
    "sableye",
    "mawile",
    "aron",
    "lairon",
    "aggron",
    "meditite",
    "medicham",
    "electrike",
    "manectric",
    "plusle",
    "minun",
    "volbeat",
    "illumise",
    "roselia",
    "gulpin",
    "swalot",
    "carvanha",
    "sharpedo",
    "wailmer",
    "wailord",
    "numel",
    "camerupt",
    "torkoal",
    "spoink",
    "grumpig",
    "spinda",
    "trapinch",
    "vibrava",
    "flygon",
    "cacnea",
    "cacturne",
    "swablu",
    "altaria",
    "zangoose",
    "seviper",
    "lunatone",
    "solrock",
    "barboach",
    "whiscash",
    "corphish",
    "crawdaunt",
    "baltoy",
    "claydol",
    "lileep",
    "cradily",
    "anorith",
    "armaldo",
    "feebas",
    "milotic",
    "castform",
    "kecleon",
    "shuppet",
    "banette",
    "duskull",
    "dusclops",
    "tropius",
    "chimecho",
    "absol",
    "wynaut",
    "snorunt",
    "glalie",
    "spheal",
    "sealeo",
    "walrein",
    "clamperl",
    "huntail",
    "gorebyss",
    "relicanth",
    "luvdisc",
    "bagon",
    "shelgon",
    "salamence",
    "beldum",
    "metang",
    "metagross",
    "regirock",
    "regice",
    "registeel",
    "latias",
    "latios",
    "kyogre",
    "groudon",
    "rayquaza",
    "jirachi",
    "deoxys",
    "turtwig",
    "grotle",
    "torterra",
    "chimchar",
    "monferno",
    "infernape",
    "piplup",
    "prinplup",
    "empoleon",
    "starly",
    "staravia",
    "staraptor",
    "bidoof",
    "bibarel",
    "kricketot",
    "kricketune",
    "shinx",
    "luxio",
    "luxray",
    "budew",
    "roserade",
    "cranidos",
    "rampardos",
    "shieldon",
    "bastiodon",
    "burmy",
    "wormadam",
    "mothim",
    "combee",
    "vespiquen",
    "pachirisu",
    "buizel",
    "floatzel",
    "cherubi",
    "cherrim",
    "shellos",
    "gastrodon",
    "ambipom",
    "drifloon",
    "drifblim",
    "buneary",
    "lopunny",
    "mismagius",
    "honchkrow",
    "glameow",
    "purugly",
    "chingling",
    "stunky",
    "skuntank",
    "bronzor",
    "bronzong",
    "bonsly",
    "happiny",
    "chatot",
    "spiritomb",
    "gible",
    "gabite",
    "garchomp",
    "munchlax",
    "riolu",
    "lucario",
    "hippopotas",
    "hippowdon",
    "skorupi",
    "drapion",
    "croagunk",
    "toxicroak",
    "carnivine",
    "finneon",
    "lumineon",
    "mantyke",
    "snover",
    "abomasnow",
    "weavile",
    "magnezone",
    "lickilicky",
    "rhyperior",
    "tangrowth",
    "electivire",
    "magmortar",
    "togekiss",
    "yanmega",
    "leafeon",
    "glaceon",
    "gliscor",
    "mamoswine",
    "gallade",
    "probopass",
    "dusknoir",
    "froslass",
    "rotom",
    "uxie",
    "mesprit",
    "azelf",
    "dialga",
    "palkia",
    "heatran",
    "regigigas",
    "giratina",
    "cresselia",
    "phione",
    "manaphy",
    "darkrai",
    "shaymin",
    "arceus",
    "victini",
    "snivy",
    "servine",
    "serperior",
    "tepig",
    "pignite",
    "emboar",
    "oshawott",
    "dewott",
    "samurott",
    "patrat",
    "watchog",
    "lillipup",
    "herdier",
    "stoutland",
    "purrloin",
    "liepard",
    "pansage",
    "simisage",
    "pansear",
    "simisear",
    "panpour",
    "simipour",
    "munna",
    "musharna",
    "pidove",
    "tranquill",
    "unfezant",
    "blitzle",
    "zebstrika",
    "roggenrola",
    "boldore",
    "gigalith",
    "woobat",
    "swoobat",
    "drilbur",
    "excadrill",
    "audino",
    "timburr",
    "gurdurr",
    "conkeldurr",
    "tympole",
    "palpitoad",
    "seismitoad",
    "throh",
    "sawk",
    "sewaddle",
    "swadloon",
    "leavanny",
    "venipede",
    "whirlipede",
    "scolipede",
    "cottonee",
    "whimsicott",
    "petilil",
    "lilligant",
    "basculin",
    "sandile",
    "krokorok",
    "krookodile",
    "darumaka",
    "darmanitan",
    "maractus",
    "dwebble",
    "crustle",
    "scraggy",
    "scrafty",
    "sigilyph",
    "yamask",
    "cofagrigus",
    "tirtouga",
    "carracosta",
    "archen",
    "archeops",
    "trubbish",
    "garbodor",
    "zorua",
    "zoroark",
    "minccino",
    "cinccino",
    "gothita",
    "gothorita",
    "gothitelle",
    "solosis",
    "duosion",
    "reuniclus",
    "ducklett",
    "swanna",
    "vanillite",
    "vanillish",
    "vanilluxe",
    "deerling",
    "sawsbuck",
    "emolga",
    "karrablast",
    "escavalier",
    "foongus",
    "amoonguss",
    "frillish",
    "jellicent",
    "alomomola",
    "joltik",
    "galvantula",
    "ferroseed",
    "ferrothorn",
    "klink",
    "klang",
    "klinklang",
    "tynamo",
    "eelektrik",
    "eelektross",
    "elgyem",
    "beheeyem",
    "litwick",
    "lampent",
    "chandelure",
    "axew",
    "fraxure",
    "haxorus",
    "cubchoo",
    "beartic",
    "cryogonal",
    "shelmet",
    "accelgor",
    "stunfisk",
    "mienfoo",
    "mienshao",
    "druddigon",
    "golett",
    "golurk",
    "pawniard",
    "bisharp",
    "bouffalant",
    "rufflet",
    "braviary",
    "vullaby",
    "mandibuzz",
    "heatmor",
    "durant",
    "deino",
    "zweilous",
    "hydreigon",
    "larvesta",
    "volcarona",
    "cobalion",
    "terrakion",
    "virizion",
    "tornadus",
    "thundurus",
    "reshiram",
    "zekrom",
    "landorus",
    "kyurem",
    "keldeo",
    "meloetta",
    "genesect",
    "chespin",
    "quilladin",
    "chesnaught",
    "fennekin",
    "braixen",
    "delphox",
    "froakie",
    "frogadier",
    "greninja",
    "bunnelby",
    "diggersby",
    "fletchling",
    "fletchinder",
    "talonflame",
    "scatterbug",
    "spewpa",
    "vivillon",
    "litleo",
    "pyroar",
    "floette",
    "florges",
    "skiddo",
    "gogoat",
    "pancham",
    "pangoro",
    "furfrou",
    "espurr",
    "meowstic",
    "honedge",
    "doublade",
    "aegislash",
    "spritzee",
    "aromatisse",
    "swirlix",
    "slurpuff",
    "inkay",
    "malamar",
    "binacle",
    "barbaracle",
    "skrelp",
    "dragalge",
    "clauncher",
    "clawitzer",
    "helioptile",
    "heliolisk",
    "tyrunt",
    "tyrantrum",
    "amaura",
    "aurorus",
    "sylveon",
    "hawlucha",
    "dedenne",
    "carbink",
    "goomy",
    "sliggoo",
    "goodra",
    "klefki",
    "phantump",
    "trevenant",
    "pumpkaboo",
    "gourgeist",
    "bergmite",
    "avalugg",
    "noibat",
    "noivern",
    "xerneas",
    "yveltal",
    "zygarde",
    "diancie",
    "hoopa",
    "volcanion",
    "rowlet",
    "dartrix",
    "decidueye",
    "litten",
    "torracat",
    "incineroar",
    "popplio",
    "brionne",
    "primarina",
    "pikipek",
    "trumbeak",
    "toucannon",
    "yungoos",
    "gumshoos",
    "grubbin",
    "charjabug",
    "vikavolt",
    "crabrawler",
    "crabominable",
    "oricorio",
    "cutiefly",
    "ribombee",
    "rockruff",
    "lycanroc",
    "wishiwashi",
    "mareanie",
    "toxapex",
    "mudbray",
    "mudsdale",
    "dewpider",
    "araquanid",
    "fomantis",
    "lurantis",
    "morelull",
    "shiinotic",
    "salandit",
    "salazzle",
    "stufful",
    "bewear",
    "bounsweet",
    "steenee",
    "tsareena",
    "comfey",
    "oranguru",
    "passimian",
    "wimpod",
    "golisopod",
    "sandygast",
    "palossand",
    "pyukumuku",
    "silvally",
    "minior",
    "komala",
    "turtonator",
    "togedemaru",
    "mimikyu",
    "bruxish",
    "drampa",
    "dhelmise",
    "cosmog",
    "cosmoem",
    "solgaleo",
    "lunala",
    "nihilego",
    "buzzwole",
    "pheromosa",
    "xurkitree",
    "celesteela",
    "kartana",
    "guzzlord",
    "necrozma",
    "magearna",
    "marshadow",
    "poipole",
    "naganadel",
    "stakataka",
    "blacephalon",
    "zeraora",
    "meltan",
    "melmetal",
]


def random_name() -> str:
    return f"{random.choice(pokemons_names)}-{datetime.now().strftime('%Y_%m_%d_%H_%M_%S')}"



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim


## TODO WR
class Network(nn.Module):
    def __init__(self, dim, hdim):
        super().__init__()
        self.up = nn.Linear(dim, dim)
        self.down = nn.Linear(dim, dim)

    def forward(self, x):
        x = self.up(x)
        x = F.silu(x)
        return self.down(x)


def log_prob(logits, labels):
    return torch.gather(logits.log_softmax(-1), dim=2, index=labels.unsqueeze(2)).squeeze(2)


dim = 10
seq = 3
batch = 1

policy = Network(dim, dim)
ref_model = Network(dim, dim)
ref_model.load_state_dict(policy.state_dict())
ref_model = ref_model.eval()

optimizer = optim.AdamW(policy.parameters(), lr=1e-2)
labels = torch.tensor([1, 2, 3]).unsqueeze(0)
inputs = torch.rand((batch, seq, dim))

for _idx in range(100):
    logits = policy(inputs)
    pi = log_prob(logits, labels).sum()
    with torch.no_grad():
        logits = ref_model(inputs)
        ref = log_prob(logits, labels).sum()

    # print(pi, ref, pi - ref)
    loss = F.sigmoid(pi - ref)
    print(loss)
    loss.backward()
    optimizer.step()
    optimizer.zero_grad()


class MoE(nn.Module):
    def __init__(self, input_dim, output_dim, num_experts, hidden_dim):
        super(MoE, self).__init__()
        self.num_experts = num_experts
        self.experts = nn.ModuleList(
            [
                nn.Sequential(
                    nn.Linear(input_dim, hidden_dim), nn.ReLU(), nn.Linear(hidden_dim, output_dim)
                )
                for _ in range(num_experts)
            ]
        )
        self.gate = nn.Linear(input_dim, num_experts)

    def forward(self, x):
        gate_outputs = F.softmax(self.gate(x), dim=-1)
        expert_outputs = torch.stack([expert(x) for expert in self.experts], dim=1)
        output = torch.sum(gate_outputs.unsqueeze(-1) * expert_outputs, dim=1)
        return output



import torch
import torch.nn as nn

from dataclasses import dataclass

from ohara.modules.mlp import SwiGLU
from ohara.modules.norm import RMSNorm
from ohara.embedings_pos.xpos import XPos


@dataclass
class Config:
    vocab_size: int = 65
    seq_len: int = 64
    d_model: int = 128
    num_heads: int = 4
    num_layers: int = 4
    dropout: float = 0.2
    multiple_of: int = 4
    bias: bool = False
    eps: float = 1e-5


# rotary embedding


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
    # [: (dim // 2)] for odd number truncation
    # torch.arange(0, dim, 2) -> 2(i-1)//d while i= 1,2,..,(d//2)

    t = torch.arange(end, device=freqs.device)
    freqs = torch.outer(t, freqs).float()  # gives diffrent angle vector

    # e^it = cos(t) + i sin(t)
    freqs_cos = torch.cos(freqs)  # real
    freqs_sin = torch.sin(freqs)  # imaginary
    return freqs_cos, freqs_sin


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
    ndim = x.dim()
    assert 1 < ndim
    assert freqs_cis.shape == (
        x.shape[1],
        x.shape[-1],
    ), f"{freqs_cis.shape=}, {(x.shape[1], x.shape[-1])=}"

    # keep 2nd (T) and last(freq) dim same else make dim 1 for freq_cis
    shape = [d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)]
    # print(shape)
    return freqs_cis.view(shape)


def apply_rope(k, q, freqs_sin, freqs_cos):
    # Idea suppose vector v = [x,y,x1,y1,...] # v.shape = dim
    # convert vetor into complex num # ie two vec one real, one imagery
    # [x,y,x1,y1,...] -> x+iy, x1+iy1
    # Multiplying by complex num == roatate vector
    # => (x + iy) * (cos + isin) -> x'+iy'
    # restack
    # x'+iy' -> [x',y',x1',y1'...]
    # you roated vector in chunks of two lfg!!!

    #  rehsape a shape (...,n )-> (..., n//2,2)
    q_cis = q.float().reshape(
        q.shape[:-1] + (-1, 2)
    )  # (B,T,nhead,C) -> (B,T,nhead,Cc,2) # Cc = C//2
    k_cis = k.float().reshape(k.shape[:-1] + (-1, 2))  # (B,T,nhead,C) -> (B,T,nhead,Cc,2)

    xq_r, xq_i = q_cis.unbind(-1)  # (B,T,nhead,Cc,2) -> ((B,T,Cc), (B,T,Cc)) split into two tuple
    xk_r, xk_i = k_cis.unbind(-1)  # (B,T,nhead,Cc,2) -> ((B,T,Cc), (B,T,Cc))

    freqs_cos = reshape_for_broadcast(freqs_cos, xq_r)  # freqs.shape = (1,T,1,Cc)
    freqs_sin = reshape_for_broadcast(freqs_cos, xq_r)

    # e+if = (a+ib) * (c+di) = (ac-bd) + i (ad+bc)
    # a = xq_r , b = xq_i
    # c = fcos , d = fsin
    # ...
    # e = (ac-bd) = xq_r * freqs_cos - xq_i * freqs_sin
    # f = (c+di)  = xq_r * freqs_sin + xq_i * freqs_cos

    xq_out_r = xq_r * freqs_cos - xq_i * freqs_sin  # (ac-bd)   # shape =  # (B,T,nhead,Cc)
    xq_out_i = xq_r * freqs_sin + xq_i * freqs_cos  # (ad+bc) * i
    xk_out_r = xk_r * freqs_cos - xk_i * freqs_sin  # (ac-bd)
    xk_out_i = xk_r * freqs_sin + xk_i * freqs_cos  # (ad+bc) * i

    # now we stack r,i -> [r,i,r2,i2]
    xq_out = torch.stack([xq_out_r, xq_out_i], dim=-1)  # (B,T,nhead,Cc,2)
    xk_out = torch.stack([xk_out_r, xk_out_i], dim=-1)  # (B,T,nhead,Cc,2)

    # flatten last two dimensions
    xq_out = xq_out.flatten(3)  # (B,T,nhead,C)
    xk_out = xk_out.flatten(3)  # (B,T,nhead,C)

    return xq_out.type_as(q), xk_out.type_as(q)


class Retation(nn.Module):
    def __init__(self, model_args: Config):
        super().__init__()
        self.d_model = model_args.d_model
        self.num_heads = model_args.num_heads
        self.head_dim = model_args.d_model // model_args.num_heads
        self.key_dim = self.d_model // self.num_heads
        self.scaling = self.key_dim**-0.5

        self.key = nn.Linear(self.d_model, self.d_model)
        self.query = nn.Linear(self.d_model, self.d_model)
        self.value = nn.Linear(self.d_model, self.d_model)
        self.gate = nn.Linear(self.d_model, self.d_model)
        self.proj = nn.Linear(self.d_model, self.d_model)

        self.norm = RMSNorm(self.head_dim, model_args.e)

        self.flash_attn = hasattr(torch.nn.functional, "scaled_dot_product_attention")

    def forward(self, x: torch.Tensor, freqs_cos, freqs_sin, mask: torch.Tensor) -> torch.Tensor:
        batch, seq_len, d_model = x.shape

        k: torch.Tensor  # type hint for lsp
        q: torch.Tensor  # ignore
        v: torch.Tensor

        k = self.key(x)
        q = self.query(x)
        v = self.value(x)
        g = self.gate(x)

        k = k * self.scaling

        k = k.view(
            seq_len, self.num_heads, self.head_dim
        )  # shape = (B, seq_len, num_heads, head_dim)
        q = q.view(seq_len, self.num_heads, self.head_dim)
        v = v.view(seq_len, self.num_heads, self.head_dim)

        q, k = apply_rope(q, k, freqs_cos, freqs_sin)

        k = k.transpose(0, 1)  # shape = (B, num_heads, seq_len, head_dim)
        q = q.transpose(0, 1)
        v = v.transpose(0, 1)

        ret_mtx = torch.matmul(q, k.transpose(2, 3))
        ret_mtx = ret_mtx / ret_mtx.detach().abs().sum(dim=-1, keepdim=True).clamp(min=1, max=5e4)
        ret_mtx = ret_mtx + mask[:, :, :seq_len, :seq_len]

        output = torch.matmul(ret_mtx, v)  # (batch, n_head, seq_len, head_dim)

        output = self.norm(output)

        output = output.transpose(1, 2).contiguous().view(batch, seq_len, d_model)

        output = output * g

        output = self.proj(output)
        return output


class Block(nn.Module):
    def __init__(self, model_args: Config):
        super().__init__()

        self.attn = Retation(model_args)
        self.ff = SwiGLU(
            dim=model_args.d_model,
            multiple_of=model_args.multiple_of,
            dropout=model_args.dropout,
        )

        self.norm1 = nn.LayerNorm(model_args.d_model)
        self.norm2 = nn.LayerNorm(model_args.d_model)

    def forward(self, x, freqs_cos, freqs_sin, mask):
        x = x + self.attn(self.norm1(x), freqs_cos, freqs_sin, mask)
        x = x + self.ff(self.norm2(x))
        return x


class RoFormer(nn.Module):
    def __init__(self, model_args: Config, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.word_emb = nn.Embedding(model_args.vocab_size, model_args.d_model)
        self.pos_emb = nn.Embedding(model_args.seq_len, model_args.d_model)

        self.layers = nn.ModuleList([Block(model_args) for _ in range(model_args.num_layers)])

        self.norm = nn.LayerNorm(model_args.d_model)
        self.vocab_proj = nn.Linear(model_args.d_model, model_args.vocab_size, bias=False)

        (cos, sin), mask = XPos(model_args.d_model, model_args.seq_len).forward(
            slen=model_args.seq_len
        )
        self.register_buffer("mask", mask, persistent=False)
        self.register_buffer("freqs_cos", cos, persistent=False)
        self.register_buffer("freqs_sin", sin, persistent=False)

    def forward(self, x):
        B, T = x.shape
        x = self.word_emb(x)
        freqs_cos = self.freqs_cos[:T]
        freqs_sin = self.freqs_sin[:T]

        for layer in self.layers:
            x = layer(x, freqs_cos, freqs_sin, self.mask)

        x = self.norm(x)
        x = self.vocab_proj(x)
        return x



import jax
import equinox as eqx


class SwiGLU(eqx.Module):
    dim: int
    hidden_dim: int
    w1: eqx.nn.Linear
    w2: eqx.nn.Linear
    w3: eqx.nn.Linear

    def __init__(self, dim, hidden, *, key):
        super().__init__()
        key_w1, key_w2, key_w3 = jax.random.split(key, 3)
        self.dim = dim
        self.hidden_dim = hidden
        self.w1 = eqx.nn.Linear(dim, hidden, key=key_w1)
        self.w2 = eqx.nn.Linear(hidden, dim, key=key_w2)  # this is order in meta's llama
        self.w3 = eqx.nn.Linear(dim, hidden, key=key_w3)

    def __call__(self, x):
        x = jax.nn.silu(jax.vmap(self.w1)(x)) * jax.vmap(self.w3)(x)
        x = jax.vmap(self.w2)(x)
        return x


if __name__ == "__main__":
    key = jax.random.PRNGKey(0)
    B, T, C = 3, 5, 256
    model = SwiGLU(dim=C, hidden=C * 4, key=key)
    x = jax.random.normal(key, (B, T, C))

    out = jax.vmap(model)(x)

    print(out.shape)



from __future__ import annotations

import torch


def svd_approx(
    W: torch.Tensor, r: int | None = None
) -> tuple[torch.Tensor, torch.Tensor, int, torch.Tensor]:
    """
    Perform SVD approximation of a matrix with an optional rank specification.

    Parameters:
    W (torch.Tensor): The input matrix to be approximated.
    r (int, optional): The rank for the approximation. If None, the full rank is used.

    Returns:
    Tuple[torch.Tensor, torch.Tensor, int, torch.Tensor]:
        A tuple containing the matrices A and B (where A @ B is the approximation of W),
        the rank used for the approximation, and the Frobenius norm of the difference between W and its approximation.
    """
    U, S, V = torch.linalg.svd(W, full_matrices=False)
    if r is None:
        r = S.size(0)
    else:
        r = min(r, S.size(0))

    A = U[:, :r]
    B = torch.diag(S[:r]) @ V[:r, :]
    approx_W = A @ B
    frobenius_norm = torch.linalg.norm(W - approx_W)
    return A, B, r, frobenius_norm



import torch
import torch.nn.functional as F


def relu_squared(x):
    return torch.relu(x) ** 2


ACT2FN = {
    "relu": F.relu,
    "silu": F.silu,
    "gelu": F.gelu,
    "tanh": torch.tanh,
    "sigmoid": torch.sigmoid,
    "relu_squared": relu_squared,
    "relu2": relu_squared,
}



from __future__ import annotations

import math

import torch
import torch.nn as nn


class LoRALinear(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        rank: int = 0,
        lora_alpha: int = 1,
        lora_dropout: float = 0.0,
        **kwargs,
    ):
        super().__init__()
        self.rank = rank
        self.lora_alpha = lora_alpha
        self.merged = False

        self.lora_dropout = nn.Dropout(p=lora_dropout) if lora_dropout > 0.0 else lambda x: x
        self.linear = torch.nn.Linear(in_features, out_features, **kwargs)

        if rank > 0:
            self.lora_A = nn.Parameter(torch.zeros((rank, in_features)))
            self.lora_B = nn.Parameter(torch.zeros((out_features, rank)))
            self.scaling = self.lora_alpha / self.rank
            self.reset_parameters()

    def reset_lora_parameters(self, lora_only=False):
        nn.init.kaiming_uniform_(self.lora_A, a=math.sqrt(5))
        nn.init.zeros_(self.lora_B)
        # funny story behind math.sqrt(5) I'll write blog later

    def reset_parameters(self, lora_only=False):
        nn.init.kaiming_normal_(self.linear.weight, a=math.sqrt(5))
        if hasattr(self, "lora_A") and hasattr(self, "lora_B"):
            self.reset_lora_parameters()

    def lora_trainable_only(self):
        self.linear.train(False)
        self.lora_A.requires_grad = True
        self.lora_B.requires_grad = True

    def merge(self):
        if not self.merged and self.rank > 0:
            self.linear.weight.data += (self.lora_B @ self.lora_A) * self.scaling
            self.merged = True

    def forward(self, x: torch.Tensor):
        pretrained = self.linear(x)
        if self.r == 0 or self.merged:
            return pretrained
        lora = (self.lora_dropout(x) @ self.lora_A.T @ self.lora_B.T) * self.scaling
        return pretrained + lora


def lora_from_linear(linear: nn.Linear, lora_alpha: int = 1, lora_dropout: float = 0.0):
    device = linear.weight.device
    dtype = linear.weight.dtype
    lora = LoRALinear(
        linear.in_features,
        linear.out_features,
        lora_alpha=lora_alpha,
        lora_dropout=lora_dropout,
    )
    lora = lora.to(device)
    lora.load_state_dict(linear.state_dict(), strict=False)  # Ensure only matching keys are loaded
    return lora.to(device).to(dtype)


def replace_with_lora(
    model: nn.Module,
    target_layer: list[str] | None = None,
    lora_alpha: int = 1,
    lora_dropout: float = 0.0,
):
    if isinstance(model, nn.Linear):
        return lora_from_linear(model, lora_alpha, lora_dropout)

    if isinstance(model, (nn.Module, nn.ModuleDict)):
        for name, module in model.named_children():
            if isinstance(module, nn.Linear) and (target_layer is None or name in target_layer):
                setattr(model, name, lora_from_linear(module, lora_alpha, lora_dropout))
            else:
                replace_with_lora(module, target_layer, lora_alpha, lora_dropout)
    return model


def mark_lora_as_trainable(model: nn.Module, target_layer: list[str] | None = None):
    # freeze hole model
    for param in model.parameters():
        param.requires_grad = False
    for module in model.modules():
        if isinstance(module, LoRALinear):
            module.lora_trainable_only()
    return model


def merge_lora(model: nn.Module, target_layer: list[str] | None = None):
    for module in model.modules():
        if isinstance(module, LoRALinear):
            module.merge()
    return model


if __name__ == "__main__":

    class Network(nn.Module):
        def __init__(self):
            super().__init__()
            self.linear = nn.Linear(2, 2)
            self.layers = nn.ModuleList([nn.Linear(2, 2) for _ in range(3)])
            self.seq = nn.Sequential(nn.Linear(2, 2), nn.Linear(2, 2))

        def forward(self, x):
            return self.linear(x)

    model = Network()
    model = replace_with_lora(model)
    print(model)



from __future__ import annotations

import os
from pathlib import Path

from transformers import AutoTokenizer
from datasets import load_dataset, DownloadMode


class DatasetPreprocessor:
    def __init__(
        self,
        dataset_name: str = "JeanKaddour/minipile",
        tokenizer_name: str = "NeelNanda/gpt-neox-tokenizer-digits",
        name:str=None,
        min_length: int = 512,
        max_length: int = 2049,
        splits: list[str] | None = None,
        revision: str | None = None,
        num_proc: int | None = None,
        output_dir: Path = Path("./data"),
        hf_cache: Path = Path("./hf_cache"),
    ):
        if splits is None:
            splits = ["train", "test"]
        self.dataset_name = dataset_name
        self.tokenizer_name = tokenizer_name
        self.name = name
        self.min_length = min_length
        self.max_length = max_length
        self.splits = splits
        self.output_dir = output_dir
        self.hf_cache = hf_cache
        self.revision = revision

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        self.tokenizer.padding_side = "right"
        self.PAD = self.tokenizer.pad_token_id
        self.length = self.tokenizer.vocab_size

        self.num_proc = num_proc if num_proc else max(os.cpu_count() - 3, 1)

    def load_and_preprocess_dataset(self, split, remove_columns=None):
        if remove_columns is None:
            remove_columns = ["text"]
        dataset = load_dataset(
            self.dataset_name,
            split=split,
            name=self.name,
            download_mode=DownloadMode.REUSE_CACHE_IF_EXISTS,
            revision=self.revision,
            cache_dir=self.hf_cache,
        )

        tokenized = (
            dataset.map(
                self.tokenize_fn,
                batched=True,
                remove_columns=remove_columns,
                num_proc=self.num_proc,
            )
            .shuffle(seed=31415)
            .filter(self.filter_fn, num_proc=self.num_proc)
        )

        return tokenized

    def tokenize_fn(self, x):
        return self.tokenizer(x["text"], max_length=self.max_length, truncation=True)

    def filter_fn(self, x):
        return len(x["input_ids"]) >= self.min_length

    def save_pre_tokenized_dataset(self, dataset, split):
        if not self.output_dir.exists():
            self.output_dir.mkdir(parents=True, exist_ok=True)
        fpath = str(
            f"{self.dataset_name.replace('/','-')}--{self.tokenizer.name_or_path.replace('/','-')}"
        )
        fpath = str(self.output_dir.joinpath(fpath).joinpath(split))
        dataset.save_to_disk(fpath)

        print(f"Dataset saved to {self.output_dir}")

    def push_pre_tokenized_dataset(self, dataset, split, hf_username):
        if not self.output_dir.exists():
            self.output_dir.mkdir(parents=True, exist_ok=True)
        fpath = str(
            f"{hf_username}/pretokenized__{self.dataset_name.replace('/','_')}__{self.tokenizer.name_or_path.replace('/','__')}"
        )
        # fpath = str(self.output_dir.joinpath(fpath).joinpath(split))
        dataset.push_to_hub(fpath, split=split)

        print(f"Dataset Pused to hf {fpath}")

    def process_and_save(self, remove_columns=None, push=False, hf_username=None):
        if remove_columns is None:
            remove_columns = ["text"]
        for split in self.splits:
            tokenized_dataset = self.load_and_preprocess_dataset(
                split, remove_columns=remove_columns
            )
            self.save_pre_tokenized_dataset(tokenized_dataset, split)
            print(f"{split} Processing and saving completed.")
            if push:
                assert hf_username is not None
                self.push_pre_tokenized_dataset(tokenized_dataset, split, hf_username)


class OpenHermesDatasetPreprocessor(DatasetPreprocessor):
    def load_and_preprocess_dataset(self, split, remove_columns=None):
        if remove_columns is None:
            remove_columns = ["text"]
        dataset = load_dataset(
            self.dataset_name,
            split=split,
            download_mode=DownloadMode.REUSE_CACHE_IF_EXISTS,
            revision=self.revision,
        )

        def map_hermes_to_chatml_format(item):
            lst = []
            for row in item["conversations"]:
                if row["from"] == "system":
                    lst.append({"role": "system", "content": row["value"]})
                if row["from"] == "gpt":
                    lst.append({"role": "assistant", "content": row["value"]})
                if row["from"] == "human":
                    lst.append({"role": "user", "content": row["value"]})
            return {"input_ids": self.tokenizer.apply_chat_template(lst)}

        tokenized = (
            dataset.map(
                map_hermes_to_chatml_format,
                num_proc=self.num_proc,
            )
            .shuffle(seed=42)
            .filter(self.filter_fn, num_proc=self.num_proc)
        )

        return tokenized

    def filter_fn(self, x):
        return len(x["input_ids"]) >= self.min_length and len(x["input_ids"]) <= self.max_length


if __name__ == "__main__":
    # Example usage
    # preprocessor = DatasetPreprocessor(splits=["test"])
    # preprocessor.process_and_save()

    preprocessor = OpenHermesDatasetPreprocessor(
        dataset_name="teknium/OpenHermes-2.5",
        tokenizer_name="philschmid/gemma-tokenizer-chatml",
        revision="f7e624a58ce3642ec50483cb6039468ee8c3c464",
        splits=["train"],
        num_proc=1,
    )

    preprocessor.process_and_save(remove_columns=None)



import torch
import torch.nn as nn

from torch import Tensor


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-5):
        super().__init__()
        """
        Paper: https://arxiv.org/abs/1910.07467
        """
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x: Tensor):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x):
        output = self._norm(x.float()).type_as(x)
        return output * self.weight
